"""
Abstract syntax tree renderer for mistletoe.
"""

import json
from mistletoe.renderers.base import BaseRenderer


class JsonRenderer(BaseRenderer):
    def render(self, token, as_string=True):
        """
        Returns the JSON string representation of the AST.

        Overrides super().render. Delegates the logic to ast_to_json.
        """
        dct = ast_to_json(token)
        if as_string:
            return json.dumps(dct, indent=2) + "\n"
        return dct

    def __getattr__(self, name):
        return lambda token: ""


def ast_to_json(token):
    """
    Recursively unrolls token attributes into dictionaries (token.children
    into lists).

    Returns:
        a dictionary of token's attributes.
    """
    node = {}
    # Python 3.6 uses [ordered dicts] [1].
    # Put in 'type' entry first to make the final tree format somewhat
    # similar to [MDAST] [2].
    #
    #   [1]: https://docs.python.org/3/whatsnew/3.6.html
    #   [2]: https://github.com/syntax-tree/mdast
    node["type"] = token.name
    node.update(token.to_dict())
    if "header" in node:
        node["header"] = ast_to_json(token.header)
    if node.get("front_matter", None) is not None:
        node["front_matter"] = ast_to_json(token.front_matter)
    if token.children is not None:
        node["children"] = [ast_to_json(child) for child in token.children]
    return node



from unittest import TestCase, mock
from mistletoe.block_tokens import Document, Heading
from mistletoe.block_tokenizer import FileWrapper
from contrib.toc_renderer import TOCRenderer


class TestTOCRenderer(TestCase):
    def test_parse_rendered_heading(self):
        rendered_heading = "<h3>some <em>text</em></h3>"
        content = TOCRenderer.parse_rendered_heading(rendered_heading)
        self.assertEqual(content, "some text")

    def test_render_heading(self):
        renderer = TOCRenderer()
        Heading.start("### some *text*\n")
        token = Heading.read(FileWrapper(["foo"]), expand_spans=True)
        renderer.render_heading(token)
        self.assertEqual(renderer._headings[0], (3, "some text"))

    def test_depth(self):
        renderer = TOCRenderer(depth=3)
        token = Document.read(["# title\n", "## heading\n", "#### heading\n"])
        renderer.render(token)
        self.assertEqual(renderer._headings, [(2, "heading")])

    def test_omit_title(self):
        renderer = TOCRenderer(omit_title=True)
        token = Document.read(["# title\n", "\n", "## heading\n"])
        renderer.render(token)
        self.assertEqual(renderer._headings, [(2, "heading")])

    def test_filter_conditions(self):
        import re

        filter_conds = [
            lambda x: re.match(r"heading", x),
            lambda x: re.match(r"foo", x),
        ]
        renderer = TOCRenderer(filter_conds=filter_conds)
        token = Document.read(
            ["# title\n", "\n", "## heading\n", "\n", "#### not heading\n"]
        )
        renderer.render(token)
        self.assertEqual(renderer._headings, [(4, "not heading")])

    @mock.patch("mistletoe.block_tokens.List")
    def test_get_toc(self, MockList):
        headings = [
            (1, "heading 1"),
            (2, "subheading 1"),
            (2, "subheading 2"),
            (3, "subsubheading 1"),
            (2, "subheading 3"),
            (1, "heading 2"),
        ]
        renderer = TOCRenderer(omit_title=False)
        renderer._headings = headings
        renderer.toc
        MockList.assert_called_with(
            [
                "- heading 1\n",
                "    - subheading 1\n",
                "    - subheading 2\n",
                "        - subsubheading 1\n",
                "    - subheading 3\n",
                "- heading 2\n",
            ]
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pathlib import Path
import sys
from importlib import import_module
from time import perf_counter


TEST_FILE = str(Path(__file__).parent.joinpath("syntax.md"))
TIMES = 1000


def benchmark(package_name):
    def decorator(func):
        def inner():
            try:
                package = import_module(package_name)
            except ImportError:
                return f"not available: '{package_name}'."

            start = perf_counter()
            for i in range(TIMES):
                func(package)
            end = perf_counter()

            return end - start

        return inner

    return decorator


@benchmark("markdown")
def run_markdown(package):
    with open(TEST_FILE, "r") as fin:
        return package.markdown(fin.read(), extensions=["extra"])


@benchmark("mistune")
def run_mistune(package):
    with open(TEST_FILE, "r") as fin:
        return package.markdown(fin.read())


@benchmark("commonmark")
def run_commonmark(package):
    with open(TEST_FILE, "r") as fin:
        return package.commonmark(fin.read())


@benchmark("mistletoe")
def run_mistletoe(package):
    with open(TEST_FILE, "r") as fin:
        return package.markdown(fin)


def run(package_name):
    print(package_name, end=": ")
    print(globals()["run_{}".format(package_name.lower())]())


def run_all(package_names):
    prompt = "Running tests with {}...".format(", ".join(package_names))
    print(prompt)
    print("=" * len(prompt))
    for package_name in package_names:
        run(package_name)


def main(*args):
    print("Test document: {}".format(TEST_FILE))
    print("Test iterations: {}".format(TIMES))
    if args[1:]:
        run_all(args[1:])
    else:
        run_all(["markdown", "mistune", "commonmark", "mistletoe"])


if __name__ == "__main__":
    main(*sys.argv)



"""Tokenize nested span tokens."""
import re

from mistletoe.span_tokens_ext import Math, Strikethrough
from mistletoe.parse_context import get_parse_context


whitespace = {" ", "\t", "\n", "\x0b", "\x0c", "\r"}
unicode_whitespace = {
    "\t",
    "\n",
    "\x0b",
    "\x0c",
    "\r",
    "\x1c",
    "\x1d",
    "\x1e",
    "\x1f",
    " ",
    "\x85",
    "\xa0",
    "\u1680",
    "\u2000",
    "\u2001",
    "\u2002",
    "\u2003",
    "\u2004",
    "\u2005",
    "\u2006",
    "\u2007",
    "\u2008",
    "\u2009",
    "\u200a",
    "\u2028",
    "\u2029",
    "\u202f",
    "\u205f",
    "\u3000",
}
punctuation = {
    "!",
    '"',
    "#",
    "$",
    "%",
    "&",
    "'",
    "(",
    ")",
    "*",
    "+",
    ",",
    "-",
    ".",
    "/",
    ":",
    ";",
    "<",
    "=",
    ">",
    "?",
    "@",
    "[",
    "\\",
    "]",
    "^",
    "_",
    "`",
    "{",
    "|",
    "}",
    "~",
}
code_pattern = re.compile(r"(?<!\\|`)(?:\\\\)*(`+)(?!`)(.+?)(?<!`)\1(?!`)", re.DOTALL)


def find_nested_tokenizer(string):
    get_parse_context().nesting_matches = {}  # reset nesting matches
    if not string:
        return []
    delimiters = []
    matches = []
    escaped = False  # escaped denotes that the last cursor position had `\`
    in_delimiter_run = None  # delimiter runs are sequences of `*` or `_`
    in_image = False
    start = 0
    i = 0

    has_math = Math in get_parse_context().span_tokens
    has_strikethrough = Strikethrough in get_parse_context().span_tokens

    # these tokens are special cases,
    # because they start and end with the same character
    # therefore, we need to re-search as we progress, to reset the opening character
    code_match = code_pattern.search(string)
    strike_match = math_match = None
    if has_strikethrough:
        strike_match = Strikethrough.pattern.search(string)
    if has_math:
        math_match = Math.pattern.search(string)

    while i < len(string):

        if strike_match is not None and i == strike_match.start():
            get_parse_context().nesting_matches.setdefault("Strikethrough", []).append(
                strike_match
            )
            strike_match = Strikethrough.pattern.search(string, i + 1)
            continue

        if code_match is not None and i == code_match.start():

            if in_delimiter_run:
                delimiters.append(Delimiter(start, i, string))
            in_delimiter_run = None

            get_parse_context().nesting_matches.setdefault("InlineCode", []).append(
                code_match
            )
            i = code_match.end()
            code_match = code_pattern.search(string, i)
            if has_strikethrough:
                strike_match = Strikethrough.pattern.search(string, i)
            if has_math:
                math_match = Math.pattern.search(string, i)
            continue

        if math_match is not None and i == math_match.start():

            if in_delimiter_run:
                delimiters.append(Delimiter(start, i, string))
            in_delimiter_run = None

            get_parse_context().nesting_matches.setdefault("Math", []).append(
                math_match
            )
            i = math_match.end()
            code_match = code_pattern.search(string, i)
            math_match = Math.pattern.search(string, i)
            if has_strikethrough:
                strike_match = Strikethrough.pattern.search(string, i)
            continue

        c = string[i]
        if c == "\\" and not escaped:
            escaped = True
            i += 1
            continue
        if in_delimiter_run is not None and (c != in_delimiter_run or escaped):
            delimiters.append(Delimiter(start, i if not escaped else i - 1, string))
            in_delimiter_run = None
        if in_delimiter_run is None and (c == "*" or c == "_") and not escaped:
            in_delimiter_run = c
            start = i
        if not escaped:
            if c == "[":
                if not in_image:
                    delimiters.append(Delimiter(i, i + 1, string))
                else:
                    delimiters.append(Delimiter(i - 1, i + 1, string))
                    in_image = False
            elif c == "!":
                in_image = True
            elif c == "]":
                i = find_link_image(string, i, delimiters, matches)
                code_match = code_pattern.search(string, i)
                if has_strikethrough:
                    strike_match = Strikethrough.pattern.search(string, i)
                if has_math:
                    math_match = Math.pattern.search(string, i)
            elif in_image:
                in_image = False
        else:
            escaped = False
        i += 1
    if in_delimiter_run:
        delimiters.append(Delimiter(start, i, string))
    process_emphasis(string, None, delimiters, matches)
    return matches


def find_link_image(string, offset, delimiters, matches):
    i = len(delimiters) - 1
    for delimiter in delimiters[::-1]:
        # found a link/image delimiter
        if delimiter.type in ("[", "!["):
            # not active, remove delimiter
            if not delimiter.active:
                delimiters.remove(delimiter)
                return offset
            match = match_link_image(string, offset, delimiter)
            # found match
            if match:
                # parse for emphasis
                process_emphasis(string, i, delimiters, matches)
                # append current match
                matches.append(match)
                # if match is a link, set all previous links to be inactive
                if delimiter.type == "[":
                    deactivate_delimiters(delimiters, i, "[")
                # shift index till end of match
                return match.end() - 1
            # no match, remove delimiter
            delimiters.remove(delimiter)
            return offset
        i -= 1
    # no link/image delimiter
    return offset


def process_emphasis(string, stack_bottom, delimiters, matches):
    star_bottom = stack_bottom
    underscore_bottom = stack_bottom
    curr_pos = next_closer(stack_bottom, delimiters)
    while curr_pos is not None:
        closer = delimiters[curr_pos]
        bottom = star_bottom if closer.type[0] == "*" else underscore_bottom
        open_pos = matching_opener(curr_pos, delimiters, bottom)
        if open_pos is not None:
            opener = delimiters[open_pos]
            n = 2 if closer.number >= 2 and opener.number >= 2 else 1
            start = opener.end - n
            end = closer.start + n
            match = MatchObj(
                start, end, (start + n, end - n, string[start + n : end - n])
            )
            match.type = "Strong" if n == 2 else "Emphasis"
            matches.append(match)
            # remove all delimiters in between
            del delimiters[open_pos + 1 : curr_pos]
            curr_pos -= curr_pos - open_pos - 1
            # remove appropriate number of chars from delimiters
            if not opener.remove(n, left=False):
                delimiters.remove(opener)
                curr_pos -= 1
            if not closer.remove(n, left=True):
                delimiters.remove(closer)
                curr_pos -= 1
            if curr_pos < 0:
                curr_pos = 0
        else:
            bottom = curr_pos - 1 if curr_pos > 1 else None
            if closer.type[0] == "*":
                star_bottom = bottom
            else:
                underscore_bottom = bottom
            if not closer.open:
                delimiters.remove(closer)
            else:
                curr_pos += 1
        curr_pos = next_closer(curr_pos, delimiters)
    del delimiters[stack_bottom:]


def match_link_image(string, offset, delimiter):
    image = delimiter.type == "!["
    start = delimiter.start
    text_start = start + delimiter.number
    text_end = offset
    text = string[text_start:text_end]
    # inline link
    if follows(string, offset, "("):
        # link destination
        match_info = match_link_dest(string, offset + 1)
        if match_info is not None:
            dest_start, dest_end, dest = match_info
            # link title
            match_info = match_link_title(string, dest_end)
            if match_info is not None:
                title_start, title_end, title = match_info
                # assert closing paren
                paren_index = shift_whitespace(string, title_end)
                if paren_index < len(string) and string[paren_index] == ")":
                    end = paren_index + 1
                    match = MatchObj(
                        start,
                        end,
                        (text_start, text_end, text),
                        (dest_start, dest_end, dest),
                        (title_start, title_end, title),
                    )
                    match.type = "Link" if not image else "Image"
                    return match
    # link definition reference
    if follows(string, offset, "["):
        # full link definition reference
        result = match_link_label(string, offset + 1)
        if result:
            match_info, (dest, title) = result
            end = match_info[1]
            match = MatchObj(
                start,
                end,
                (text_start, text_end, text),
                (-1, -1, dest),
                (-1, -1, title),
            )
            match.type = "Link" if not image else "Image"
            return match
        ref = is_link_label(text)
        if ref:
            # compact link definition reference
            if follows(string, offset + 1, "]"):
                dest, title = ref
                end = offset + 3
                match = MatchObj(
                    start,
                    end,
                    (text_start, text_end, text),
                    (-1, -1, dest),
                    (-1, -1, title),
                )
                match.type = "Link" if not image else "Image"
                return match
        return None
    # shortcut link definition reference
    ref = is_link_label(text)
    if ref:
        dest, title = ref
        end = offset + 1
        match = MatchObj(
            start, end, (text_start, text_end, text), (-1, -1, dest), (-1, -1, title)
        )
        match.type = "Link" if not image else "Image"
        return match
    return None


def match_link_dest(string, offset):
    offset = shift_whitespace(string, offset + 1)
    if string[offset] == "<":
        escaped = False
        for i, c in enumerate(string[offset + 1 :], start=offset + 1):
            if c == "\\" and not escaped:
                escaped = True
            elif c == " " or c == "\n" or (c == "<" and not escaped):
                return None
            elif c == ">" and not escaped:
                return offset, i + 1, string[offset + 1 : i]
            elif escaped:
                escaped = False
        return None
    else:
        escaped = False
        count = 1
        for i, c in enumerate(string[offset:], start=offset):
            if c == "\\" and not escaped:
                escaped = True
            elif c in whitespace:
                return offset, i, string[offset:i]
            elif not escaped:
                if c == "(":
                    count += 1
                elif c == ")":
                    count -= 1
            elif is_control_char(c):
                return None
            elif escaped:
                escaped = False
            if count == 0:
                return offset, i, string[offset:i]
        return None


def match_link_title(string, offset):
    offset = shift_whitespace(string, offset)
    if string[offset] == ")":
        return offset, offset, ""
    if string[offset] == '"':
        closing = '"'
    elif string[offset] == "'":
        closing = "'"
    elif string[offset] == "(":
        closing = ")"
    else:
        return None
    escaped = False
    for i, c in enumerate(string[offset + 1 :], start=offset + 1):
        if c == "\\" and not escaped:
            escaped = True
        elif c == closing and not escaped:
            return offset, i + 1, string[offset + 1 : i]
        elif escaped:
            escaped = False
    return None


def match_link_label(string, offset):
    start = -1
    end = -1
    escaped = False
    for i, c in enumerate(string[offset:], start=offset):
        if c == "\\" and not escaped:
            escaped = True
        elif c == "[" and not escaped:
            if start == -1:
                start = i
            else:
                return None
        elif c == "]" and not escaped:
            end = i
            label = string[start + 1 : end]
            match_info = start, end + 1, label
            if label.strip() != "":
                link_definitions = get_parse_context().link_definitions
                ref = link_definitions.get(normalize_label(label), None)
                if ref is not None:
                    return match_info, ref
                return None
            return None
        elif escaped:
            escaped = False
    return None


def is_link_label(text):
    escaped = False
    for c in text:
        if c == "\\" and not escaped:
            escaped = True
        elif (c == "[" or c == "]") and not escaped:
            return None
        elif escaped:
            escaped = False
    if text.strip() != "":
        link_definitions = get_parse_context().link_definitions
        return link_definitions.get(normalize_label(text), None)
    return None


def normalize_label(text):
    return " ".join(text.split()).casefold()


def next_closer(curr_pos, delimiters):
    for i, delimiter in enumerate(delimiters[curr_pos:], start=curr_pos or 0):
        if hasattr(delimiter, "close") and delimiter.close:
            return i
    return None


def matching_opener(curr_pos, delimiters, bottom):
    if curr_pos > 0:
        curr_delimiter = delimiters[curr_pos]
        index = curr_pos - 1
        for delimiter in delimiters[curr_pos - 1 : bottom : -1]:
            if (
                hasattr(delimiter, "open")
                and delimiter.open
                and delimiter.closed_by(curr_delimiter)
            ):
                return index
            index -= 1
    return None


def is_opener(start, end, string):
    if string[start] == "*":
        return is_left_delimiter(start, end, string)
    is_right = is_right_delimiter(start, end, string)
    return is_left_delimiter(start, end, string) and (
        not is_right or (is_right and preceded_by(start, string, punctuation))
    )


def is_closer(start, end, string):
    if string[start] == "*":
        return is_right_delimiter(start, end, string)
    is_left = is_left_delimiter(start, end, string)
    return is_right_delimiter(start, end, string) and (
        not is_left or (is_left and succeeded_by(end, string, punctuation))
    )


def is_left_delimiter(start, end, string):
    return not succeeded_by(end, string, unicode_whitespace) and (
        not succeeded_by(end, string, punctuation)
        or preceded_by(start, string, punctuation)
        or preceded_by(start, string, unicode_whitespace)
    )


def is_right_delimiter(start, end, string):
    return not preceded_by(start, string, unicode_whitespace) and (
        not preceded_by(start, string, punctuation)
        or succeeded_by(end, string, unicode_whitespace)
        or succeeded_by(end, string, punctuation)
    )


def preceded_by(start, string, charset):
    preceding_char = string[start - 1] if start > 0 else " "
    return preceding_char in charset


def succeeded_by(end, string, charset):
    succeeding_char = string[end] if end < len(string) else " "
    return succeeding_char in charset


def is_control_char(char):
    return ord(char) < 32 or ord(char) == 127


def follows(string, index, char):
    return index + 1 < len(string) and string[index + 1] == char


def shift_whitespace(string, index):
    for i, c in enumerate(string[index:], start=index):
        if c not in whitespace:
            return i
    return len(string)


def deactivate_delimiters(delimiters, index, delimiter_type):
    for delimiter in delimiters[:index]:
        if delimiter.type == delimiter_type:
            delimiter.active = False


class Delimiter:
    def __init__(self, start, end, string):
        self.type = string[start:end]
        self.number = end - start
        self.active = True
        self.start = start
        self.end = end
        if self.type.startswith(("*", "_")):
            self.open = is_opener(start, end, string)
            self.close = is_closer(start, end, string)

    def remove(self, n, left=True):
        if self.number - n == 0:
            return False
        if left:
            self.start = self.start + n
            self.number = self.end - self.start
            self.type = self.type[n:]
            return True
        self.end = self.end - n
        self.number = self.end - self.start
        self.type = self.type[:n]
        return True

    def closed_by(self, other):
        return not (
            self.type[0] != other.type[0]
            or (self.open and self.close or other.open and other.close)
            and (self.number + other.number) % 3 == 0
        )

    def __repr__(self):
        if not self.type.startswith(("*", "_")):
            return "<Delimiter type={} active={}>".format(repr(self.type), self.active)
        return "<Delimiter type={} active={} open={} close={}>".format(
            repr(self.type), self.active, self.open, self.close
        )


class MatchObj:
    def __init__(self, start, end, *fields):
        self._start = start
        self._end = end
        self.fields = fields

    def start(self, n=0):
        if n == 0:
            return self._start
        return self.fields[n - 1][0]

    def end(self, n=0):
        if n == 0:
            return self._end
        return self.fields[n - 1][1]

    def group(self, n=0):
        if n == 0:
            return "".join([field[2] for field in self.fields])
        return self.fields[n - 1][2]

    def __repr__(self):
        return "<MatchObj fields={} start={} end={}>".format(
            self.fields, self._start, self._end
        )



# Copyright 2018 Tile, Inc.  All Rights Reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase
from mistletoe.span_tokenizer import tokenize_span
from contrib.jira_renderer import JIRARenderer
import random
import string


class TestJIRARenderer(TestCase):
    def setUp(self):
        self.renderer = JIRARenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def genRandomString(self, n, hasWhitespace=False):
        source = string.ascii_letters + string.digits
        if hasWhitespace:
            source = source + " \t"

        result = "".join(random.SystemRandom().choice(source) for _ in range(n))
        return result

    def textFormatTest(self, inputTemplate, outputTemplate):
        input = self.genRandomString(80, False)
        token = next(iter(tokenize_span(inputTemplate.format(input))))
        expected = outputTemplate.format(input)
        actual = self.renderer.render(token)
        self.assertEqual(expected, actual)

    def test_render_strong(self):
        self.textFormatTest("**a{}**", "*a{}*")

    def test_render_emphasis(self):
        self.textFormatTest("*a{}*", "_a{}_")

    def test_render_inline_code(self):
        self.textFormatTest("`a{}b`", "{{{{a{}b}}}}")

    def test_render_strikethrough(self):
        self.textFormatTest("-{}-", "-{}-")

    def test_render_image(self):
        token = next(iter(tokenize_span("![image](foo.jpg)")))
        expected = "!foo.jpg!"
        actual = self.renderer.render(token)
        self.assertEqual(expected, actual)

    def test_render_link_definition_image(self):
        # token = next(tokenize_span('![image]\n\n[image]: foo.jpg'))
        # expected = '!foo.jpg!'
        # actual = self.renderer.render(token)
        # self.assertEqual(expected, actual)
        pass

    def test_render_link(self):
        url = "http://{0}.{1}.{2}".format(
            self.genRandomString(5), self.genRandomString(5), self.genRandomString(3)
        )
        body = self.genRandomString(80, True)
        token = next(iter(tokenize_span("[{body}]({url})".format(url=url, body=body))))
        expected = "[{body}|{url}]".format(url=url, body=body)
        actual = self.renderer.render(token)
        self.assertEqual(expected, actual)

    def test_render_link_definition(self):
        pass

    def test_render_auto_link(self):
        url = "http://{0}.{1}.{2}".format(
            self.genRandomString(5), self.genRandomString(5), self.genRandomString(3)
        )
        token = next(iter(tokenize_span("<{url}>".format(url=url))))
        expected = "[{url}]".format(url=url)
        actual = self.renderer.render(token)
        self.assertEqual(expected, actual)

    def test_render_escape_sequence(self):
        pass

    def test_render_html_span(self):
        pass

    def test_render_heading(self):
        pass

    def test_render_quote(self):
        pass

    def test_render_paragraph(self):
        pass

    def test_render_block_code(self):
        pass

    def test_render_list(self):
        pass

    def test_render_list_item(self):
        pass

    def test_render_inner(self):
        pass

    def test_render_table(self):
        pass

    def test_render_table_row(self):
        pass

    def test_render_table_cell(self):
        pass

    def test_render_thematic_break(self):
        pass

    def test_render_html_block(self):
        pass

    def test_render_document(self):
        pass



from collections import namedtuple
import json
from typing import List, Optional, Pattern

import attr


WalkItem = namedtuple("WalkItem", ["node", "parent", "depth"])


class Token:
    """Base class of all mistletoe tokens."""

    def __getattr__(self, name):
        # ensure certain attributes are always available
        if name == "children":
            return None
        if name == "content":
            return ""
        raise AttributeError(name)

    @property
    def name(self) -> str:
        """Return the name of the element."""
        return type(self).__name__

    def __contains__(self, text: str):
        """Return is text is contained in the element or its ancestors."""
        if self.children is None:
            return text in self.content
        return any(text in child for child in self.children)

    def __repr__(self):
        """A base represent method, that can be overriden for more complex elements."""
        info = []
        if self.children is not None:
            info.append("children={}".format(len(self.children)))
        return "{}({})".format(self.name, ",".join(info))

    def to_dict(self) -> dict:
        """Convert instatiated attributes to a dict"""
        try:
            return attr.asdict(self, recurse=False)
        except attr.exceptions.NotAnAttrsClassError:
            return self.__dict__

    def walk(
        self,
        tokens: Optional[List[str]] = None,
        depth: Optional[int] = None,
        include_self: bool = False,
    ) -> WalkItem:
        """Traverse the syntax tree, recursively yielding children.

        :param elements: filter children by certain token names.
        :param depth: The depth to recurse into the tree.
        :param include_self: whether to first yield this element.

        :yield: A container for an element, its parent and depth

        """
        current_depth = 0
        if include_self:
            yield WalkItem(self, None, current_depth)
        next_children = [(self, c) for c in self.children or []]
        if self.name == "Table" and getattr(self, "header", None) is not None:
            # table headers row
            next_children.append((self, self.header))
        while next_children and (depth is None or current_depth > depth):
            current_depth += 1
            new_children = []
            for idx, (parent, child) in enumerate(next_children):
                if tokens is None or child.name in tokens:
                    yield WalkItem(child, parent, current_depth)
                new_children.extend([(child, c) for c in child.children or []])
                if child.name == "Table" and getattr(child, "header", None) is not None:
                    # table headers row
                    new_children.append((child, child.header))

            next_children = new_children


class TokenEncoder(json.JSONEncoder):
    """A JSON encoder for mistletoe tokens."""

    def default(self, obj):
        if isinstance(obj, SpanContainer):
            return list(obj.expand())
        if isinstance(obj, Token):
            return {obj.name: obj.to_dict()}
        return super().default(obj)


def serialize_tokens(tokens, as_dict=False):
    """Serialize one or more tokens, to a JSON representation."""
    string = json.dumps(tokens, cls=TokenEncoder)
    if as_dict:
        return json.loads(string)
    return string


class SpanContainer:
    """This is a container for inline span text.

    We use it in order to delay the assessment of span text, when parsing a document,
    so that all link definitions can be gathered first.
    After the initial block parse, we walk through the document
    and replace these span containers with the actual span tokens
    (see `block_tokenizer.tokenize_main`).
    """

    def __init__(self, text):
        self.text = text

    def expand(self):
        from mistletoe.span_tokenizer import tokenize_span

        return tokenize_span(self.text)

    def __iter__(self):
        for _ in []:
            yield

    def __len__(self):
        return 0


class BlockToken(Token):
    """Base class for block-level tokens. Recursively parse inner tokens.

    Naming conventions:

    * lines denotes a list of (possibly unparsed) input lines,
      and is commonly used as the argument name for constructors.

    * BlockToken.children is a list with all the inner tokens
      (thus if a token has children attribute, it is not a leaf node; if a token
      calls tokenize_span, it is the boundary between
      span-level tokens and block-level tokens);

    * BlockToken.start takes a line from the document as argument,
      and returns a boolean representing whether that line marks the start
      of the current token. Every subclass of BlockToken must define a
      start function (see block_tokenizer.tokenize).

    * BlockToken.read takes the rest of the lines in the document as an
      iterator (including the start line), and consumes all the lines
      that should be read into this token.

      Default to stop at an empty line.

      Note that `BlockToken.read` returns a token (or None).

      If BlockToken.read returns None, the read result is ignored,
      but the token class is responsible for resetting the iterator
      to a previous state. See `block_tokenizer.FileWrapper.anchor`,
      `block_tokenizer.FileWrapper.reset`.

    """

    @classmethod
    def start(cls, line: str) -> bool:
        """Takes a line from the document as argument, and
        returns a boolean representing whether that line marks the start
        of the current token. Every subclass of BlockToken must define a
        start function (see `block_tokenizer.tokenize_main`).
        """
        raise NotImplementedError

    @classmethod
    def read(cls, lines) -> Optional[Token]:
        """takes the rest of the lines in the document as an
        iterator (including the start line), and consumes all the lines
        that should be read into this token.

        The default is to stop at an empty line.
        """
        line_buffer = [next(lines)]
        for line in lines:
            if line == "\n":
                break
            line_buffer.append(line)
        return line_buffer


class SpanToken(Token):
    """Base class for span-level tokens.

    :cvar pattern: regex pattern to search for.
    :cvar parse_inner: whether to do a nested parse of the content
    :cvar parse_group: the group within the pattern match corresponding to the content
    :cvar precedence: Alter the relative order by which the span token is assessed.

    :ivar content: raw string content of the token
    :ivar children: list of child tokens
    """

    pattern = None
    parse_inner = True
    parse_group = 1
    precedence = 5

    def __init__(
        self, *, content: Optional[str] = None, children: Optional[list] = None
    ):
        if content is not None:
            self.content = content
        if children is not None:
            self.children = children

    @classmethod
    def read(cls, match: Pattern):
        """Take a pattern match and return the instatiated token."""
        if not cls.parse_inner:
            return cls(content=match.group(cls.parse_group))
        return cls()

    @classmethod
    def find(cls, string: str):
        """Find all tokens, matching a pattern in the given string"""
        if cls.pattern is not None:
            return cls.pattern.finditer(string)
        return []



"""``attrs`` class wrapper to create a docstring from the metadata."""
import textwrap
import warnings

import attr


def type_to_sphinx(typ, field_name):
    """Convert a type object to a string acceptable by Sphinx."""
    # TODO better implementation of type_to_sphinx
    if typ is None:
        field_type = "Any"
        # warnings.warn(f'Field "{field_name}" has no declared type.')
    elif getattr(typ, "__module__", None) == "typing":
        field_type = str(typ).replace("typing.", "")
        if field_type.startswith("Union["):
            field_type = field_type[6:-1]
    elif (
        hasattr(typ, "__name__")
        and getattr(typ, "__module__", "builtins") != "builtins"
    ):
        field_type = f"{typ.__module__}.{typ.__name__}"
    elif hasattr(typ, "__name__"):
        field_type = typ.__name__
    else:
        warnings.warn(f'Field "{field_name}" type could not be converted.')
        field_type = "Any"
    return field_type


def autodoc(attrs_class):
    """Decorate an :class:`attr.s` class to update its docstring with attributes.

    The docstring will include the :class:`attributes <attr.ib>`,
    and emits warnings when partially undocumented attributes are found.

    If the class contains any attributes,
    they are documented as *constructor parameters* in a *Parameters* section,
    added at the end of the class' docstring.
    In order to extract the most information per parameter description, this
    decorator assumes that every attribute

    * Typed (either via the ``type`` argument, or via `PEP 526`_ type-annotations), and
    * Contains a metadata dictionary with the keys

        help:
            The string that describes the attribute,
            and will go in the description of the corresponding parameter.

    Default values are, of course, optional, and if provided,
    the parameter's description will inform this.
    Private attributes will have their names rendered correctly (no leading underscore),
    and non-private attributes help-text will be complemented with a reminder that
    the parameter's value can be later accessed via an attribute with that name.

    .. warning::

        Currently, there is no support for indicating that arguments are keyword-only,
        hashable, or validated.
        These and other exceptional conditions should be informed in the docstring.

    .. _PEP 526: https://www.python.org/dev/peps/pep-0526/
    """
    # noqa: D202

    def fix_indent(docstring):
        lines = docstring.split("\n")

        if len(lines) > 1:
            for i in range(1, len(lines)):
                if lines[i]:
                    break
            parts = ["\n".join(lines[:i]), textwrap.dedent("\n".join(lines[i:]))]
            result = "\n".join(p for p in parts if p)
            return result
        else:
            return lines[0]

    def param_doc(field):
        field_fn = f"{attrs_class.__module__}.{attrs_class.__name__}.{field.name}"
        help_complement = "  "

        name = field.name[1:] if field.name[0] == "_" else field.name
        field_type = type_to_sphinx(field.type, field_fn)
        type_line = f":type {name}: {field_type}"

        if field.default is not attr.NOTHING:
            if isinstance(field.default, str):
                help_complement += f'(default: "{field.default}")'
            else:
                help_complement += f"(default: {field.default})"

        if "doc" in field.metadata:
            description = textwrap.indent(
                textwrap.fill(
                    text=(field.metadata["doc"] + " " + help_complement).strip(),
                    width=88,
                ),
                prefix=" " * 4,
            )
        else:
            description = help_complement

        descript_line = f":param {name}: {description}"

        return descript_line + "\n" + type_line

    if attr.fields(attrs_class):
        params_section = "\n\n".join(
            param_doc(field) for field in attr.fields(attrs_class)
        )
    else:
        params_section = ""

    if attrs_class.__doc__ and params_section:
        attrs_class.__doc__ = (
            f"{fix_indent(attrs_class.__doc__).rstrip()}\n\n{params_section}"
        )
    elif params_section:
        attrs_class.__doc__ = params_section

    return attrs_class



import pytest

from mistletoe.span_tokenizer import tokenize_span
from mistletoe.span_tokens import CoreTokens, HTMLSpan
from mistletoe.span_tokens_ext import Math
from mistletoe.base_elements import serialize_tokens
from mistletoe.parse_context import get_parse_context


@pytest.mark.parametrize(
    "name,source",
    [
        ("star", "**some text**"),
        ("underline", "__some text__"),
        ("multi", "**one** **two**"),
    ],
)
def test_strong(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_strong_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [("star", "*some text*"), ("underline", "_some text_"), ("multi", "*one* *two*")],
)
def test_emphasis(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_emphasis_{name}",
    )


@pytest.mark.parametrize(
    "name,source", [("basic", "`some text`"), ("multi", "`one` `two`")]
)
def test_inline_code(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_inline_code_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", "~~some text~~"),
        ("multi", "~~one~~ ~~two~~"),
        ("nesting", "~~*some text*~~"),
    ],
)
def test_strikethrough(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_strikethrough_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", "[name 1](target1)"),
        ("multi", "[n1](t1) & [n2](t2)"),
        ("children", "[![alt](src)](target)"),
    ],
)
def test_link(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_link_{name}",
    )


@pytest.mark.parametrize("name,source", [("basic", "<ftp://foo.com>")])
def test_auto_link(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_auto_link_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", "![alt](link)"),
        ("with_title", '![alt](link "title")'),
        ("no_alt", "![](link)"),
    ],
)
def test_image(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_image_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [("single_star", "\\*"), ("emphasis", "some \\*text*"), ("in_code", "`a \\* b`")],
)
def test_escape_sequence(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_escape_sequence_{name}",
    )


@pytest.mark.parametrize("name,source", [("basic", "some text")])
def test_raw_text(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_raw_text_{name}",
    )


@pytest.mark.parametrize("name,source", [("basic", "  \n")])
def test_line_break(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_line_break_{name}",
    )


@pytest.mark.parametrize("name,source", [("basic", "<p>some text</p>")])
def test_html_span(name, source, data_regression):
    get_parse_context().span_tokens.insert(1, HTMLSpan)
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_html_span_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", "$a$"),
        ("contains_special_chars", "$a`{_*-%$"),
        ("preceding_special_chars", "{_*-%`$a$"),
        ("multiple", "$a$ $b$"),
        ("escaped_opening", "\\$a $b$"),
        ("no_closing", "$a"),
        ("internal_emphasis", "$*a*$"),
        ("external_emphasis", "*$a$*"),
        ("multi-line", "$$a\nc\nb$$"),
        ("dollar_in_code", "a `$` `$x=1$` renders $x=1$."),
        ("in_link_content", "[$a$](link)"),
        ("in_link_target", "[a]($b$)"),
        ("in_image", "![$a$]($b$)"),
    ],
)
def test_math_span(name, source, data_regression):
    _span_tokens = get_parse_context().span_tokens
    _span_tokens.insert(_span_tokens.index(CoreTokens) + 1, Math)
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_math_span_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("emph_in_strong", "**with some *emphasis* text**"),
        ("strong_in_emph", "*with some **strong** text*"),
        ("star_underline", "*__some text__*"),
        ("underline_star", "__*some text*__"),
        ("emph_in_link", "[*some text*](link)"),
        ("link_in_emph", "*[*some text*](link)*"),
        ("star_in_inline_code", "`*a*`"),
        ("underscore_in_inline_code", "`_a_`"),
        ("inline_code_in_star", "*`a`*"),
        ("inline_code_in_underscore", "_`a`_"),
        ("auto_link_plus_code", "`<`abc>  <abc`>`"),
        ("strikethrough_plus_code", "`~~`abc ~~abc~~ `~~`"),
    ],
)
def test_nested(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_span(source), as_dict=True),
        basename=f"test_nested_{name}",
    )



"""
Make mistletoe easier to import.
"""

__version__ = "0.9.4a2"
__all__ = [
    "renderers",
    "base_elements",
    "block_tokens",
    "block_tokenizer",
    "span_tokens",
    "span_tokenizer",
]

from mistletoe.block_tokens import Document
from mistletoe.renderers.base import BaseRenderer  # noqa: F401
from mistletoe.renderers.html import HTMLRenderer


def markdown(iterable, renderer=HTMLRenderer):
    """
    Output HTML with default settings.
    Enables inline and block-level HTML tags.
    """
    with renderer() as renderer:
        return renderer.render(Document.read(iterable))



"""
Built-in block-level token classes.
"""
import re
from typing import Optional, Tuple
from typing import List as ListType

import attr

import mistletoe.block_tokenizer as tokenizer
from mistletoe import span_tokens
from mistletoe.nested_tokenizer import (
    follows,
    shift_whitespace,
    whitespace,
    is_control_char,
    normalize_label,
)
from mistletoe.parse_context import get_parse_context
from mistletoe.base_elements import Token, BlockToken, SpanContainer
from mistletoe.attr_doc import autodoc


"""
Tokens to be included in the parsing process, in the order specified.
"""
__all__ = [
    "BlockCode",
    "Heading",
    "Quote",
    "CodeFence",
    "ThematicBreak",
    "List",
    "LinkDefinition",
    "Paragraph",
]


@autodoc
@attr.s(slots=True, kw_only=True)
class FrontMatter(BlockToken):
    """Front matter YAML block.

    ::

        ---
        a: b
        c: d
        ---

    NOTE: The content of the block should be valid YAML,
    but its parsing (and hence syntax testing) is deferred to the renderers.
    This is so that, given 'bad' YAML,
    the rest of the of document will still be parsed,
    and then the renderers can apply there own error reporting.

    Not included in the parsing process, but called by `Document.read`.
    """

    content: str = attr.ib(
        repr=False, metadata={"doc": "Source text (should be valid YAML)"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    @classmethod
    def start(cls, line: str) -> bool:
        return False

    @classmethod
    def read(cls, lines):
        assert lines and lines[0].startswith("---")
        end_line = None
        for i, line in enumerate(lines[1:]):
            if line.startswith("---"):
                end_line = i + 2
                break
        # TODO raise/report error if closing block not found
        if end_line is None:
            end_line = len(lines)

        return cls(content="".join(lines[1 : end_line - 1]), position=(0, end_line))


@autodoc
@attr.s(slots=True, kw_only=True)
class Document(BlockToken):
    """Document container."""

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    link_definitions: dict = attr.ib(
        repr=lambda d: str(len(d)), metadata={"doc": "Mapping of keys to (url, title)"}
    )
    front_matter: Optional[FrontMatter] = attr.ib(
        default=None, metadata={"doc": "Front matter YAML block"}
    )

    @classmethod
    def read(
        cls,
        lines,
        start_line: int = 0,
        reset_definitions=True,
        store_definitions=False,
        front_matter=False,
    ):
        """Read a document

        :param lines:  Lines or string to parse
        :param start_line: The initial line (used for nested parsing)
        :param reset_definitions: remove any previously stored link_definitions
        :param store_definitions: store LinkDefinitions or ignore them
        :param front_matter: search for an initial YAML block front matter block
            (note this is not strictly CommonMark compliant)
        """
        if isinstance(lines, str):
            lines = lines.splitlines(keepends=True)
        lines = [line if line.endswith("\n") else "{}\n".format(line) for line in lines]
        # reset link definitions
        if reset_definitions:
            get_parse_context().link_definitions = {}

        front_matter_token = None
        if front_matter and lines and lines[0].startswith("---"):
            front_matter_token = FrontMatter.read(lines)
            start_line += front_matter_token.position[1]
            lines = lines[front_matter_token.position[1] :]

        children = tokenizer.tokenize_main(
            lines, start_line=start_line, store_definitions=store_definitions
        )
        return cls(
            children=children,
            front_matter=front_matter_token,
            link_definitions=get_parse_context().link_definitions,
        )


@autodoc
@attr.s(slots=True, kw_only=True)
class Heading(BlockToken):
    """Heading token. (["### some heading ###\\n"])

    Boundary between span-level and block-level tokens.
    """

    level: int = attr.ib(metadata={"doc": "Heading level"})
    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    pattern = re.compile(r" {0,3}(#{1,6})(?:\n|\s+?(.*?)(?:\n|\s+?#+\s*?$))")

    @classmethod
    def start(cls, line):
        match_obj = cls.pattern.match(line)
        if match_obj is None:
            return False
        cls.level = len(match_obj.group(1))
        cls.content = (match_obj.group(2) or "").strip()
        if set(cls.content) == {"#"}:
            cls.content = ""
        return True

    @classmethod
    def read(cls, lines, expand_spans=False):
        next(lines)
        children = SpanContainer(cls.content)
        if expand_spans:
            children = children.expand()
        return cls(
            level=cls.level, children=children, position=(lines.lineno, lines.lineno)
        )


@autodoc
@attr.s(slots=True, kw_only=True)
class SetextHeading(BlockToken):
    """Setext headings.

    Not included in the parsing process, but returned by `Paragraph.read`.
    """

    level: int = attr.ib(metadata={"doc": "Heading level"})
    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    @classmethod
    def start(cls, line):
        raise NotImplementedError()

    @classmethod
    def read(cls, lines):
        raise NotImplementedError()


@autodoc
@attr.s(slots=True, kw_only=True)
class Quote(BlockToken):
    """Quote token. (`["> # heading\\n", "> paragraph\\n"]`)."""

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    @staticmethod
    def start(line):
        stripped = line.lstrip(" ")
        if len(line) - len(stripped) > 3:
            return False
        return stripped.startswith(">")

    @classmethod
    def transition(cls, next_line):
        return (
            next_line is None
            or next_line.strip() == ""
            or Heading.start(next_line)
            or CodeFence.start(next_line)
            or ThematicBreak.start(next_line)
            or List.start(next_line)
        )

    @classmethod
    def read(cls, lines):
        # first line
        start_line = lines.lineno + 1
        line = cls.convert_leading_tabs(next(lines).lstrip()).split(">", 1)[1]
        if len(line) > 0 and line[0] == " ":
            line = line[1:]
        line_buffer = [line]

        # set booleans
        in_code_fence = CodeFence.start(line)
        in_block_code = BlockCode.start(line)
        blank_line = line.strip() == ""

        # loop
        next_line = lines.peek()
        while not cls.transition(next_line):
            stripped = cls.convert_leading_tabs(next_line.lstrip())
            prepend = 0
            if stripped[0] == ">":
                # has leader, not lazy continuation
                prepend += 1
                if stripped[1] == " ":
                    prepend += 1
                stripped = stripped[prepend:]
                in_code_fence = CodeFence.start(stripped)
                in_block_code = BlockCode.start(stripped)
                blank_line = stripped.strip() == ""
                line_buffer.append(stripped)
            elif in_code_fence or in_block_code or blank_line:
                # not paragraph continuation text
                break
            else:
                # lazy continuation, preserve whitespace
                line_buffer.append(next_line)
            next(lines)
            next_line = lines.peek()

        # block level tokens are parsed here, so that link_definitions
        # in quotes can be recognized before span-level tokenizing.
        Paragraph.parse_setext = False
        try:
            child_tokens = tokenizer.tokenize_block(line_buffer, start_line=start_line)
        finally:
            Paragraph.parse_setext = True
        return cls(children=child_tokens, position=(start_line, lines.lineno))

    @staticmethod
    def convert_leading_tabs(string):
        string = string.replace(">\t", "   ", 1)
        count = 0
        for i, c in enumerate(string):
            if c == "\t":
                count += 4
            elif c == " ":
                count += 1
            else:
                break
        if i == 0:
            return string
        return ">" + " " * count + string[i:]


@autodoc
@attr.s(slots=True, kw_only=True)
class Paragraph(BlockToken):
    """Paragraph token. (`["some\\n", "continuous\\n", "lines\\n"]`)

    Boundary between span-level and block-level tokens.
    """

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    _setext_pattern = re.compile(r" {0,3}(=|-)+ *$")
    parse_setext = True  # can be disabled by Quote

    @staticmethod
    def start(line):
        return line.strip() != ""

    @classmethod
    def is_setext_heading(cls, line):
        return cls._setext_pattern.match(line)

    @classmethod
    def transition(cls, next_line):
        return (
            next_line is None
            or next_line.strip() == ""
            or Heading.start(next_line)
            or CodeFence.start(next_line)
            or Quote.start(next_line)
        )

    @classmethod
    def read(cls, lines, expand_spans=False):
        line_buffer = [next(lines)]
        start_line = lines.lineno
        next_line = lines.peek()
        while not cls.transition(next_line):
            # check if next_line starts List
            list_pair = ListItem.parse_marker(next_line)
            if len(next_line) - len(next_line.lstrip()) < 4 and list_pair is not None:
                prepend, leader = list_pair
                # non-empty list item
                if next_line[:prepend].endswith(" "):
                    # unordered list, or ordered list starting from 1
                    if not leader[:-1].isdigit() or leader[:-1] == "1":
                        break

            # check if next_line starts HTMLBlock other than type 7
            html_block = HTMLBlock.start(next_line)
            if html_block and html_block != 7:
                break

            # check if we see a setext underline
            if cls.parse_setext and cls.is_setext_heading(next_line):
                line_buffer.append(next(lines))
                level = 1 if line_buffer.pop().lstrip().startswith("=") else 2
                children = SpanContainer(
                    "\n".join([line.strip() for line in line_buffer])
                )
                if expand_spans:
                    children = children.expand()
                return SetextHeading(
                    children=children, level=level, position=(start_line, lines.lineno)
                )

            # check if we have a ThematicBreak (has to be after setext)
            if ThematicBreak.start(next_line):
                break

            # no other tokens, we're good
            line_buffer.append(next(lines))
            next_line = lines.peek()

        content = "".join([line.lstrip() for line in line_buffer]).strip()
        children = SpanContainer(content)
        if expand_spans:
            children = children.expand()
        return cls(children=children, position=(start_line, lines.lineno))


@autodoc
@attr.s(slots=True, kw_only=True)
class BlockCode(BlockToken):
    """Indented code."""

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    language: str = attr.ib(
        default="", metadata={"doc": "The code language (for sytax highlighting)"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    @staticmethod
    def start(line):
        return line.replace("\t", "    ", 1).startswith("    ")

    @classmethod
    def read(cls, lines):
        start_line = lines.lineno
        line_buffer = []
        for line in lines:
            if line.strip() == "":
                line_buffer.append(line.lstrip(" ") if len(line) < 5 else line[4:])
                continue
            if not line.replace("\t", "    ", 1).startswith("    "):
                lines.backstep()
                break
            line_buffer.append(cls.strip(line))

        children = (span_tokens.RawText("".join(line_buffer).strip("\n") + "\n"),)

        return cls(children=children, language="", position=(start_line, lines.lineno))

    @staticmethod
    def strip(string):
        count = 0
        for i, c in enumerate(string):
            if c == "\t":
                return string[i + 1 :]
            elif c == " ":
                count += 1
            else:
                break
            if count == 4:
                return string[i + 1 :]
        return string


@autodoc
@attr.s(slots=True, kw_only=True)
class CodeFence(BlockToken):
    """Code fence. (["```sh\\n", "rm -rf /", ..., "```"])

    Boundary between span-level and block-level tokens.
    """

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    language: str = attr.ib(
        default="", metadata={"doc": "The code language (for sytax highlighting)"}
    )
    arguments: str = attr.ib(
        default="", metadata={"doc": "Any string occuring after the language"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    pattern = re.compile(r"^( {0,3})((?:`|~){3,}) *([^`~\s]*) *([^`~]*)$")
    _open_info = None

    @classmethod
    def start(cls, line):
        match_obj = cls.pattern.match(line)
        if not match_obj:
            return False
        prepend, leader, lang, arguments = match_obj.groups()
        if leader[0] in lang or leader[0] in line[match_obj.end() :]:
            return False
        cls._open_info = len(prepend), leader, lang, arguments
        return True

    @classmethod
    def read(cls, lines):
        start_line = lines.lineno + 1
        next(lines)
        line_buffer = []
        for line in lines:
            stripped_line = line.lstrip(" ")
            diff = len(line) - len(stripped_line)
            if (
                stripped_line.startswith(cls._open_info[1])
                and len(stripped_line.split(maxsplit=1)) == 1
                and diff < 4
            ):
                break
            if diff > cls._open_info[0]:
                stripped_line = " " * (diff - cls._open_info[0]) + stripped_line
            line_buffer.append(stripped_line)

        language = span_tokens.EscapeSequence.strip(cls._open_info[2])
        arg_lines = cls._open_info[3].splitlines() or [""]
        arguments = span_tokens.EscapeSequence.strip(arg_lines[0])
        children = (span_tokens.RawText("".join(line_buffer)),)

        return cls(
            children=children,
            language=language,
            arguments=arguments,
            position=(start_line, lines.lineno),
        )


@autodoc
@attr.s(slots=True, kw_only=True)
class List(BlockToken):
    """List token (unordered or ordered)"""

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    loose: bool = attr.ib(
        metadata={"doc": "Whether list items are separated by blank lines"}
    )
    start_at: Optional[int] = attr.ib(
        metadata={"doc": "None if unordered, starting number if ordered."}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    _pattern = re.compile(r" {0,3}(?:\d{0,9}[.)]|[+\-*])(?:[ \t]*$|[ \t]+)")

    @classmethod
    def start(cls, line):
        return cls._pattern.match(line)

    @classmethod
    def read(cls, lines):
        start_line = lines.lineno
        leader = None
        next_marker = None
        children = []
        while True:
            item = ListItem.read(lines, next_marker)
            next_marker = item.next_marker
            item_leader = item.leader
            if leader is None:
                leader = item_leader
            elif not cls.same_marker_type(leader, item_leader):
                lines.reset()
                break
            children.append(item)
            if next_marker is None:
                break

        if children:
            # Only consider the last list item loose if there's more than one element
            last_parse_buffer = children[-1]
            last_parse_buffer.loose = (
                len(last_parse_buffer.children) > 1 and last_parse_buffer.loose
            )

        loose = any(item.loose for item in children)
        leader = children[0].leader
        start = None
        if len(leader) != 1:
            start = int(leader[:-1])
        return cls(
            children=children,
            loose=loose,
            start_at=start,
            position=(start_line, lines.lineno),
        )

    @staticmethod
    def same_marker_type(leader, other):
        if len(leader) == 1:
            return leader == other
        return (
            leader[:-1].isdigit() and other[:-1].isdigit() and leader[-1] == other[-1]
        )


@autodoc
@attr.s(slots=True, kw_only=True)
class ListItem(BlockToken):
    """List items.

    Not included in the parsing process, but called by List.
    """

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    loose: bool = attr.ib(
        metadata={"doc": "Whether list items are separated by blank lines"}
    )
    leader: str = attr.ib(metadata={"doc": "The prefix number or bullet point."})
    prepend = attr.ib(metadata={"doc": ""})
    next_marker = attr.ib(metadata={"doc": ""})
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    _pattern = re.compile(r"\s*(\d{0,9}[.)]|[+\-*])(\s*$|\s+)")

    @staticmethod
    def in_continuation(line, prepend):
        return line.strip() == "" or len(line) - len(line.lstrip()) >= prepend

    @staticmethod
    def transition(next_line):
        return (
            Heading.start(next_line)
            or Quote.start(next_line)
            or CodeFence.start(next_line)
            or ThematicBreak.start(next_line)
        )

    @classmethod
    def parse_marker(cls, line):
        """
        Returns a pair (prepend, leader) if the line has a valid leader.
        """
        match_obj = cls._pattern.match(line)
        if match_obj is None:
            return None  # no valid leader
        leader = match_obj.group(1)
        content = match_obj.group(0).replace(leader + "\t", leader + "   ", 1)
        # reassign prepend and leader
        prepend = len(content)
        if prepend == len(line.rstrip("\n")):
            prepend = match_obj.end(1) + 1
        else:
            spaces = match_obj.group(2)
            if spaces.startswith("\t"):
                spaces = spaces.replace("\t", "   ", 1)
            spaces = spaces.replace("\t", "    ")
            n_spaces = len(spaces)
            if n_spaces > 4:
                prepend = match_obj.end(1) + 1
        return prepend, leader

    @classmethod
    def read(cls, lines, prev_marker=None):
        next_marker = None
        lines.anchor()
        prepend = -1
        leader = None
        start_line = lines.lineno
        line_buffer = []

        # first line
        line = next(lines)
        prepend, leader = prev_marker if prev_marker else cls.parse_marker(line)
        line = line.replace(leader + "\t", leader + "   ", 1).replace("\t", "    ")
        empty_first_line = line[prepend:].strip() == ""
        if not empty_first_line:
            line_buffer.append(line[prepend:])
        next_line = lines.peek()
        if empty_first_line and next_line is not None and next_line.strip() == "":
            child_tokens = tokenizer.tokenize_block(
                [next(lines)], start_line=lines.lineno
            )
            next_line = lines.peek()
            if next_line is not None:
                marker_info = cls.parse_marker(next_line)
                if marker_info is not None:
                    next_marker = marker_info
            return cls(
                children=child_tokens,
                loose=child_tokens.loose,
                prepend=prepend,
                leader=leader,
                next_marker=next_marker,
                position=(start_line, lines.lineno),
            )

        # loop
        newline = 0
        while True:
            # no more lines
            if next_line is None:
                # strip off newlines
                if newline:
                    lines.backstep()
                    del line_buffer[-newline:]
                break
            next_line = next_line.replace("\t", "    ")
            # not in continuation
            if not cls.in_continuation(next_line, prepend):
                # directly followed by another token
                if cls.transition(next_line):
                    if newline:
                        lines.backstep()
                        del line_buffer[-newline:]
                    break
                # next_line is a new list item
                marker_info = cls.parse_marker(next_line)
                if marker_info is not None:
                    next_marker = marker_info
                    break
                # not another item, has newlines -> not continuation
                if newline:
                    lines.backstep()
                    del line_buffer[-newline:]
                    break
            next(lines)
            line = next_line
            stripped = line.lstrip(" ")
            diff = len(line) - len(stripped)
            if diff > prepend:
                stripped = " " * (diff - prepend) + stripped
            line_buffer.append(stripped)
            newline = newline + 1 if next_line.strip() == "" else 0
            next_line = lines.peek()

        child_tokens = tokenizer.tokenize_block(line_buffer, start_line=start_line)

        return cls(
            children=child_tokens,
            loose=child_tokens.loose,
            prepend=prepend,
            leader=leader,
            next_marker=next_marker,
            position=(start_line, lines.lineno),
        )


@autodoc
@attr.s(slots=True, kw_only=True)
class LinkDefinition(BlockToken):
    """LinkDefinition token: `[ref]: url "title"`"""

    definitions: list = attr.ib(metadata={"doc": "list of (label, dest, title)"})
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    label_pattern = re.compile(r"[ \n]{0,3}\[(.+?)\]", re.DOTALL)

    @classmethod
    def start(cls, line):
        return line.lstrip().startswith("[")

    @classmethod
    def read(cls, lines):
        line_buffer = []
        start_line = lines.lineno + 1
        next_line = lines.peek()
        while next_line is not None and next_line.strip() != "":
            line_buffer.append(next(lines))
            next_line = lines.peek()
        string = "".join(line_buffer)
        offset = 0
        matches = []
        while offset < len(string) - 1:
            match_info = cls.match_reference(lines, string, offset)
            if match_info is None:
                break
            offset, match = match_info
            matches.append(match)
        cls.append_link_definitions(matches)
        return (
            cls(position=(start_line, lines.lineno), definitions=matches)
            if matches
            else None
        )

    @classmethod
    def match_reference(cls, lines, string, offset):
        match_info = cls.match_link_label(string, offset)
        if not match_info:
            cls.backtrack(lines, string, offset)
            return None
        _, label_end, label = match_info

        if not follows(string, label_end - 1, ":"):
            cls.backtrack(lines, string, offset)
            return None

        match_info = cls.match_link_dest(string, label_end)
        if not match_info:
            cls.backtrack(lines, string, offset)
            return None
        _, dest_end, dest = match_info

        match_info = cls.match_link_title(string, dest_end)
        if not match_info:
            cls.backtrack(lines, string, dest_end)
            return None
        _, title_end, title = match_info

        return title_end, (label, dest, title)

    @classmethod
    def match_link_label(cls, string, offset):
        start = -1
        end = -1
        escaped = False
        for i, c in enumerate(string[offset:], start=offset):
            if c == "\\" and not escaped:
                escaped = True
            elif c == "[" and not escaped:
                if start == -1:
                    start = i
                else:
                    return None
            elif c == "]" and not escaped:
                end = i
                label = string[start + 1 : end]
                if label.strip() != "":
                    return start, end + 1, label
                return None
            elif escaped:
                escaped = False
        return None

    @classmethod
    def match_link_dest(cls, string, offset):
        offset = shift_whitespace(string, offset + 1)
        if offset == len(string):
            return None
        if string[offset] == "<":
            escaped = False
            for i, c in enumerate(string[offset + 1 :], start=offset + 1):
                if c == "\\" and not escaped:
                    escaped = True
                elif c == " " or c == "\n" or (c == "<" and not escaped):
                    return None
                elif c == ">" and not escaped:
                    return offset, i + 1, string[offset + 1 : i]
                elif escaped:
                    escaped = False
            return None
        else:
            escaped = False
            count = 0
            for i, c in enumerate(string[offset:], start=offset):
                if c == "\\" and not escaped:
                    escaped = True
                elif c in whitespace:
                    break
                elif not escaped:
                    if c == "(":
                        count += 1
                    elif c == ")":
                        count -= 1
                elif is_control_char(c):
                    return None
                elif escaped:
                    escaped = False
            if count != 0:
                return None
            return offset, i, string[offset:i]

    @classmethod
    def match_link_title(cls, string, offset):
        new_offset = shift_whitespace(string, offset)
        if (
            new_offset == len(string)
            or "\n" in string[offset:new_offset]
            and string[new_offset] == "["
        ):
            return offset, new_offset, ""
        if string[new_offset] == '"':
            closing = '"'
        elif string[new_offset] == "'":
            closing = "'"
        elif string[new_offset] == "(":
            closing = ")"
        elif "\n" in string[offset:new_offset]:
            return offset, offset, ""
        else:
            return None
        offset = new_offset
        escaped = False
        for i, c in enumerate(string[offset + 1 :], start=offset + 1):
            if c == "\\" and not escaped:
                escaped = True
            elif c == closing and not escaped:
                new_offset = shift_whitespace(string, i + 1)
                if "\n" not in string[i + 1 : new_offset]:
                    return None
                return offset, new_offset, string[offset + 1 : i]
            elif escaped:
                escaped = False
        return None

    @staticmethod
    def append_link_definitions(matches):
        for key, dest, title in matches:
            key = normalize_label(key)
            dest = span_tokens.EscapeSequence.strip(dest.strip())
            title = span_tokens.EscapeSequence.strip(title)
            link_definitions = get_parse_context().link_definitions
            if key not in link_definitions:
                link_definitions[key] = dest, title

    @staticmethod
    def backtrack(lines, string, offset):
        lines._index -= string[offset + 1 :].count("\n")


@autodoc
@attr.s(slots=True, kw_only=True)
class ThematicBreak(BlockToken):
    """Thematic break token (a.k.a. horizontal rule.)"""

    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    _pattern = re.compile(r" {0,3}(?:([-_*])\s*?)(?:\1\s*?){2,}$")

    @classmethod
    def start(cls, line):
        return cls._pattern.match(line)

    @classmethod
    def read(cls, lines):
        next(lines)
        return cls(position=(lines.lineno, lines.lineno))


@autodoc
@attr.s(slots=True, kw_only=True)
class HTMLBlock(BlockToken):
    """Block-level HTML token."""

    content: str = attr.ib(
        repr=False, metadata={"doc": "literal strings rendered as-is"}
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    _end_cond = None
    multiblock = re.compile(r"<(script|pre|style)[ >\n]")
    predefined = re.compile(r"<\/?(.+?)(?:\/?>|[ \n])")
    custom_tag = re.compile(
        r"(?:" + "|".join((span_tokens._open_tag, span_tokens._closing_tag)) + r")\s*$"
    )

    @classmethod
    def start(cls, line):
        stripped = line.lstrip()
        if len(line) - len(stripped) >= 4:
            return False
        # rule 1: <pre>, <script> or <style> tags, allow newlines in block
        match_obj = cls.multiblock.match(stripped)
        if match_obj is not None:
            cls._end_cond = "</{}>".format(match_obj.group(1).casefold())
            return 1
        # rule 2: html comment tags, allow newlines in block
        if stripped.startswith("<!--"):
            cls._end_cond = "-->"
            return 2
        # rule 3: tags that starts with <?, allow newlines in block
        if stripped.startswith("<?"):
            cls._end_cond = "?>"
            return 3
        # rule 4: tags that starts with <!, allow newlines in block
        if stripped.startswith("<!") and stripped[2].isupper():
            cls._end_cond = ">"
            return 4
        # rule 5: CDATA declaration, allow newlines in block
        if stripped.startswith("<![CDATA["):
            cls._end_cond = "]]>"
            return 5
        # rule 6: predefined tags (see html_token._tags), read until newline
        match_obj = cls.predefined.match(stripped)
        if match_obj is not None and match_obj.group(1).casefold() in span_tokens._tags:
            cls._end_cond = None
            return 6
        # rule 7: custom tags, read until newline
        match_obj = cls.custom_tag.match(stripped)
        if match_obj is not None:
            cls._end_cond = None
            return 7
        return False

    @classmethod
    def read(cls, lines):
        # note: stop condition can trigger on the starting line
        start_line = lines.lineno
        line_buffer = []
        for line in lines:
            line_buffer.append(line)
            if cls._end_cond is not None:
                if cls._end_cond in line.casefold():
                    break
            elif line.strip() == "":
                line_buffer.pop()
                break
        return cls(
            content="".join(line_buffer).rstrip("\n"),
            position=(start_line, lines.lineno),
        )



import pytest

from mistletoe import block_tokens, block_tokens_ext
from mistletoe.base_elements import serialize_tokens
from mistletoe.parse_context import get_parse_context
from mistletoe.block_tokenizer import tokenize_main


@pytest.mark.parametrize(
    "name,source",
    [
        ("match", ["### heading 3\n"]),
        ("enclosing_hashes", ["# heading 3 #####  \n"]),
        ("too_many_hashes", ["####### paragraph\n"]),
        ("heading_in_paragraph", ["foo\n", "# heading\n", "bar\n"]),
    ],
)
def test_atx_heading(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_atx_heading_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("match", ["some heading\n", "---\n"]),
        ("multiline", ["some\n", "heading\n", "---\n", "\n", "foobar\n"]),
    ],
)
def test_setext_heading(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_setext_heading_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("match", ["> line 1\n", "> line 2\n"]),
        ("lazy_continuation", ["> line 1\n", "line 2\n"]),
    ],
)
def test_quote(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_quote_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("with_ticks", ["```sh\n", "rm dir\n", "mkdir test\n", "```\n"]),
        ("with_tildas", ["~~~sh\n", "rm dir\n", "mkdir test\n", "~~~\n"]),
        (
            "mixed_tick_tilda",
            ["~~~markdown\n", "```sh\n", "some code\n", "```\n", "~~~\n"],
        ),
        ("lazy_continuation", ["```sh\n", "rm dir\n", "\n", "mkdir test\n", "```\n"]),
        ("no_wrapping_newlines", ["```\n", "hey", "```\n", "paragraph\n"]),
        ("unclosed", ["```\n", "hey"]),
    ],
)
def test_fenced_code(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_fenced_code_{name}",
    )


@pytest.mark.parametrize(
    "name,source", [("match", ["    rm dir\n", "    mkdir test\n"])]
)
def test_block_code(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_block_code_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", ["some\n", "continuous\n", "lines\n"]),
        ("encapsulate_code_fence", ["this\n", "```\n", "is some\n", "```\n", "code\n"]),
    ],
)
def test_paragraph(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_paragraph_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        (
            "allowed_markers",
            [
                "- foo\n",
                "*    bar\n",
                " + baz\n",
                "1. item 1\n",
                "2) item 2\n",
                "123456789. item x\n",
            ],
        ),
        (
            "non_markers",
            ["> foo\n", "1item 1\n", "2| item 2\n", "1234567890. item x\n"],
        ),
        ("before_block_code", [" -    foo\n", "   bar\n", "\n", "          baz\n"]),
        ("sub_list", ["- foo\n", "  - bar\n"]),
        ("deep_list", ["- foo\n", "  - bar\n", "    - baz\n"]),
        ("loose_list", ["- foo\n", "  ~~~\n", "  bar\n", "  \n", "  baz\n" "  ~~~\n"]),
        ("tight_list", ["- foo\n", "\n", "# bar\n"]),
    ],
)
def test_list_item(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_list_item_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("markers", ["- foo\n", "* bar\n", "1. baz\n", "2) spam\n"]),
        ("sub_list", ["- foo\n", "  + bar\n"]),
    ],
)
def test_list(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_list_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", ["[key 1]: value1\n", "[key 2]: value2\n"]),
        ("title", ['[key 1]: value1 "title"\n']),
    ],
)
def test_link_definitions(name, source, data_regression):
    tree = serialize_tokens(tokenize_main(source), as_dict=True)
    data_regression.check(
        {"tree": tree, "link_definitions": get_parse_context().link_definitions},
        basename=f"test_link_definitions_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [("dash", ["---\n"]), ("star", ["* * *\n"]), ("underscore", ["_    _    _\n"])],
)
def test_thematic_break(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_thematic_break_{name}",
    )


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", ["# heading\n", "\n", "paragraph\n", "with\n", "`code`\n"]),
        ("link_defs", ["[key 1]: value1\n", "[key 2]: value2\n"]),
        ("auto_splitlines", "some\ncontinual\nlines\n"),
        ("front_matter_ignore", ["---\n", "a: 1\n", "---\n"]),
    ],
)
def test_doc_read(name, source, data_regression):
    data_regression.check(
        serialize_tokens(block_tokens.Document.read(source), as_dict=True),
        basename=f"test_doc_read_{name}",
    )


@pytest.mark.parametrize("name,source", [("basic", ["---\n", "a: 1\n", "---\n"])])
def test_doc_read_with_front_matter(name, source, data_regression):
    data_regression.check(
        serialize_tokens(
            block_tokens.Document.read(source, front_matter=True), as_dict=True
        ),
        basename=f"test_doc_read_with_front_matter_{name}",
    )


@pytest.mark.parametrize(
    "name,source", [("basic", ["[key 1]: value1\n", "[key 2]: value2\n"])]
)
def test_doc_read_store_link_defs(name, source, data_regression):
    data_regression.check(
        serialize_tokens(
            block_tokens.Document.read(source, store_definitions=True), as_dict=True
        ),
        basename=f"test_doc_read_store_link_defs_{name}",
    )


def test_table_parse_align():
    assert block_tokens_ext.Table.parse_align(":------") is None
    assert block_tokens_ext.Table.parse_align(":-----:") == 0
    assert block_tokens_ext.Table.parse_align("------:") == 1


def test_table_parse_delimiter():
    delimiters = list(
        block_tokens_ext.Table.split_delimiter("| :--- | :---: | ---:|\n")
    )
    assert delimiters == [":---", ":---:", "---:"]


@pytest.mark.parametrize(
    "name,source",
    [
        (
            "basic",
            [
                "| header 1 | header 2 | header 3 |\n",
                "| --- | --- | --- |\n",
                "| cell 1 | cell 2 | cell 3 |\n",
                "| more 1 | more 2 | more 3 |\n",
            ],
        ),
        (
            "aligned",
            ["header 1 | header 2\n", "    ---: | :---\n", "  cell 1 | cell 2\n"],
        ),
        ("not_table", ["not header 1 | not header 2\n", "foo | bar\n"]),
    ],
)
def test_table(name, source, data_regression):
    data_regression.check(
        serialize_tokens(tokenize_main(source), as_dict=True),
        basename=f"test_table_{name}",
    )


@pytest.mark.parametrize(
    "name,source,row_align",
    [
        ("basic", "| cell 1 | cell 2 |\n", None),
        ("no_outside", "cell 1 | cell 2\n", None),
        ("short", "| cell 1 |\n", [None, None]),
    ],
)
def test_table_row(name, source, row_align, data_regression):
    row = block_tokens_ext.TableRow.read(source, row_align=row_align)
    data_regression.check(
        serialize_tokens(row, as_dict=True), basename=f"test_table_row_{name}"
    )


def test_table_cell(data_regression):
    token = block_tokens_ext.TableCell.read("cell 2")
    data_regression.check(serialize_tokens(token, as_dict=True))



import os

from mistletoe import markdown

PATH = os.path.dirname(__file__)


def test_syntax(file_regression):
    with open(os.path.join(PATH, "syntax.md")) as handle:
        file_regression.check(markdown(handle.read()), extension=".html")


def test_jquery(file_regression):
    with open(os.path.join(PATH, "jquery.md")) as handle:
        file_regression.check(markdown(handle.read()), extension=".html")



from textwrap import dedent

from mistletoe import Document


def test_walk():
    doc = Document.read(
        dedent(
            """\
        a **b**

        c [*d*](link)
        """
        )
    )
    tree = [
        (t.node.name, t.parent.name if t.parent else None, t.depth)
        for t in doc.walk(include_self=True)
    ]
    assert tree == [
        ("Document", None, 0),
        ("Paragraph", "Document", 1),
        ("Paragraph", "Document", 1),
        ("RawText", "Paragraph", 2),
        ("Strong", "Paragraph", 2),
        ("RawText", "Paragraph", 2),
        ("Link", "Paragraph", 2),
        ("RawText", "Strong", 3),
        ("Emphasis", "Link", 3),
        ("RawText", "Emphasis", 4),
    ]



import unittest
from mistletoe import Document
from contrib.mathjax import MathJaxRenderer


class TestMathJaxRenderer(unittest.TestCase):
    mathjax_src = (
        '<script src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.0/MathJax.js'
        '?config=TeX-MML-AM_CHTML"></script>\n'
    )

    def test_render_html(self):
        with MathJaxRenderer() as renderer:
            token = Document.read(["# heading 1\n", "paragraph\n"])
            output = renderer.render(token)
            target = "<h1>heading 1</h1>\n<p>paragraph</p>\n"
            target += self.mathjax_src
            self.assertEqual(output, target)

    def test_render_math(self):
        with MathJaxRenderer() as renderer:
            raw = ["# heading 1\n", "$$paragraph$$\n", "with $ math $\n"]
            token = Document.read(raw)
            output = renderer.render(token)
            target = "<h1>heading 1</h1>\n<p>$$paragraph$$\nwith $$ math $$</p>\n"
            target += self.mathjax_src
            self.assertEqual(output, target)



from mistletoe.base_elements import serialize_tokens
from mistletoe.span_tokenizer import tokenize_span
from contrib.github_wiki import GithubWikiRenderer


def test_parse(data_regression):
    with GithubWikiRenderer():
        source = "text with [[wiki | target]]"
        data_regression.check(serialize_tokens(tokenize_span(source), as_dict=True))


def test_parse_with_children(data_regression):
    with GithubWikiRenderer():
        source = "[[*alt*|link]]"
        data_regression.check(serialize_tokens(tokenize_span(source), as_dict=True))


def test_render(file_regression):
    with GithubWikiRenderer() as renderer:
        token = tokenize_span("[[wiki|target]]")[0]
        file_regression.check(renderer.render(token), extension=".html")



"""
Base class for renderers.
"""

from itertools import chain
import re
import sys

from mistletoe import block_tokens, block_tokens_ext, span_tokens, span_tokens_ext
from mistletoe.parse_context import ParseContext, get_parse_context, set_parse_context


class BaseRenderer:
    """
    Base class for renderers.

    All renderers should ...

    * define all render functions specified in `self.render_map`;
    * be a context manager (by inheriting `__enter__` and `__exit__`);

    Custom renderers could ...

    * add additional tokens into the parsing process by passing custom
      tokens to `super().__init__()`;
    * add additional render functions by appending to self.render_map;

    :Usage:

    Suppose SomeRenderer inherits BaseRenderer, and fin is the input file.
    The syntax looks something like this::

        >>> from mistletoe import Document
        >>> from some_renderer import SomeRenderer
        >>> with SomeRenderer() as renderer:
        ...     rendered = renderer.render(Document.read(fin))

    See mistletoe.renderers.html for an implementation example.

    :Naming conventions:

    * The keys of `self.render_map` should exactly match the class
      name of tokens;
    * Render function names should be of form: `render_` + the
      "snake-case" form of token's class name.

    :param render_map: maps tokens to their corresponding render functions.
    :type render_map: dict
    :param parse_context: container for the instatiated tokens,
        and other global parsing variables. These will be re-instatiated on `__enter__`

    """

    default_block_tokens = (
        block_tokens.HTMLBlock,
        block_tokens.BlockCode,
        block_tokens.Heading,
        block_tokens.Quote,
        block_tokens.CodeFence,
        block_tokens.ThematicBreak,
        block_tokens.List,
        block_tokens_ext.Table,
        block_tokens.LinkDefinition,
        block_tokens.Paragraph,
    )

    default_span_tokens = (
        span_tokens.EscapeSequence,
        span_tokens.HTMLSpan,
        span_tokens.AutoLink,
        span_tokens.CoreTokens,
        span_tokens_ext.Strikethrough,
        span_tokens.InlineCode,
        span_tokens.LineBreak,
        span_tokens.RawText,
    )

    _parse_name = re.compile(r"([A-Z][a-z]+|[A-Z]+(?![a-z]))")

    def __init__(self, find_blocks=None, find_spans=None):
        """Initialise the renderer

        :param find_blocks: override the default block tokens (classes or class paths)
        :param find_spans: override the default span tokens (classes or class paths)
        """
        self.parse_context = ParseContext(
            find_blocks or self.default_block_tokens,
            find_spans or self.default_span_tokens,
        )
        set_parse_context(self.parse_context)

        self.render_map = self.get_default_render_map()
        for token in chain(
            self.parse_context.block_tokens, self.parse_context.span_tokens
        ):
            if token.__name__ not in self.render_map:
                render_func = getattr(self, self._cls_to_func(token.__name__))
                self.render_map[token.__name__] = render_func

        self.link_definitions = {}

    def get_default_render_map(self):
        """Return the default map of token names to methods."""
        return {
            "Strong": self.render_strong,
            "Emphasis": self.render_emphasis,
            "InlineCode": self.render_inline_code,
            "RawText": self.render_raw_text,
            "Strikethrough": self.render_strikethrough,
            "Image": self.render_image,
            "Link": self.render_link,
            "AutoLink": self.render_auto_link,
            "EscapeSequence": self.render_escape_sequence,
            "Heading": self.render_heading,
            "SetextHeading": self.render_setext_heading,
            "Quote": self.render_quote,
            "Paragraph": self.render_paragraph,
            "CodeFence": self.render_code_fence,
            "BlockCode": self.render_block_code,
            "List": self.render_list,
            "ListItem": self.render_list_item,
            "Table": self.render_table,
            "TableRow": self.render_table_row,
            "TableCell": self.render_table_cell,
            "ThematicBreak": self.render_thematic_break,
            "LineBreak": self.render_line_break,
            "Document": self.render_document,
            "LinkDefinition": self.render_link_definition,
        }

    def render(self, token):
        """
        Grabs the class name from input token and finds its corresponding
        render function.

        Basically a janky way to do polymorphism.

        Arguments:
            token: whose __class__.__name__ is in self.render_map.
        """
        return self.render_map[token.__class__.__name__](token)

    def render_inner(self, token):
        """
        Recursively renders child tokens. Joins the rendered
        strings with no space in between.

        If newlines / spaces are needed between tokens, add them
        in their respective templates, or override this function
        in the renderer subclass, so that whitespace won't seem to
        appear magically for anyone reading your program.

        :param token: a branch node who has children attribute.
        """
        return "".join(map(self.render, token.children or []))

    def __enter__(self):
        """
        Make renderer classes into context managers.
        """
        set_parse_context(self.parse_context)
        return self

    def __exit__(self, exception_type, exception_val, traceback):
        """
        Make renderer classes into context managers.
        """
        get_parse_context(reset=True)

    @classmethod
    def _cls_to_func(cls, cls_name):
        snake = "_".join(map(str.lower, cls._parse_name.findall(cls_name)))
        return "render_{}".format(snake)

    @staticmethod
    def _tokens_from_module(module):
        """
        Helper method; takes a module and returns a list of all token classes
        specified in `module.__all__`. Useful when custom tokens are defined in a
        separate module.
        """
        return [getattr(module, name) for name in module.__all__]

    def render_raw_text(self, token):
        """
        Default render method for RawText. Simply return token.content.
        """
        return token.content

    def render_setext_heading(self, token):
        """
        Default render method for SetextHeader. Simply parse to render_header.
        """
        return self.render_heading(token)

    def render_code_fence(self, token):
        """
        Default render method for CodeFence. Simply parse to render_block_code.
        """
        return self.render_block_code(token)

    def __getattr__(self, name):
        """
        Provides a default render method for all tokens.

        Any token without a custom render method will simply be rendered by
        self.render_inner.

        If name does not start with 'render_', raise AttributeError as normal,
        for less magic during debugging.

        This method would only be called if the attribute requested has not
        been defined. Defined attributes will not be overridden.

        I still think this is heavy wizardry.
        Let me know if you would like this method removed.
        """
        if not name.startswith("render_"):
            msg = "{cls} object has no attribute {name}".format(
                cls=type(self).__name__, name=name
            )
            raise AttributeError(msg).with_traceback(sys.exc_info()[2])
        return self.render_inner



import json
import os

import pytest

from mistletoe import Document
from mistletoe.renderers.html import HTMLRenderer

with open(os.path.join(os.path.dirname(__file__), "commonmark.json"), "r") as fin:
    tests = json.load(fin)


@pytest.mark.parametrize("entry", tests)
def test_commonmark(entry):
    test_case = entry["markdown"].splitlines(keepends=True)
    with HTMLRenderer() as renderer:
        output = renderer.render(Document.read(test_case))
    assert entry["html"] == output



"""Extended span tokens, that are not part of the CommonMark spec."""
import re
from mistletoe.base_elements import SpanToken
from mistletoe.parse_context import get_parse_context


__all__ = ["Math", "Strikethrough"]


class Strikethrough(SpanToken):
    """Strikethrough tokens: `~~some text~~`.

    Must be parsed after `CoreTokens`.
    """

    pattern = re.compile(r"(?<!\\)(?:\\\\)*~~(.+?)~~", re.DOTALL)

    @classmethod
    def find(cls, string):
        matches = get_parse_context().nesting_matches.pop("Strikethrough", [])
        return matches


class Math(SpanToken):
    """Dollar Math tokens (single or double): `$a=1$`.

    Must be parsed after `CoreTokens`.
    """

    pattern = re.compile(r"(?<!\\)(?:\\\\)*(\${1,2})([^\$]+?)\1")
    parse_inner = False
    parse_group = 0

    @classmethod
    def find(cls, string):
        matches = get_parse_context().nesting_matches.pop("Math", [])
        return matches



"""
Extended block tokens, that are not part of the CommonMark spec.
"""
from itertools import zip_longest
import re
from typing import Optional, Tuple
from typing import List as ListType

import attr

from mistletoe.attr_doc import autodoc
from mistletoe.base_elements import Token, BlockToken, SpanContainer


__all__ = ["TableCell", "TableRow", "Table"]


@autodoc
@attr.s(slots=True, kw_only=True)
class TableCell(BlockToken):
    """Table cell token.

    Boundary between span-level and block-level tokens.
    """

    children: ListType[Token] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    align: Optional[int] = attr.ib(
        metadata={
            "doc": "align options for the cell (left=None (default), center=0, right=1)"
        }
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    @classmethod
    def read(cls, content, align=None, expand_spans=False, lineno=0):
        children = SpanContainer(content)
        if expand_spans:
            children = children.expand()
        return cls(children=children, align=align, position=(lineno, lineno))


@autodoc
@attr.s(slots=True, kw_only=True)
class TableRow(BlockToken):
    """Table row token."""

    children: ListType[TableCell] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    row_align: list = attr.ib(
        metadata={
            "doc": "align options for columns (left=None (default), center=0, right=1)"
        }
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    @classmethod
    def read(cls, line, row_align=None, lineno=0):
        row_align = row_align or [None]
        cells = filter(None, line.strip().split("|"))
        children = [
            TableCell.read(cell.strip() if cell else "", align, lineno=lineno)
            for cell, align in zip_longest(cells, row_align)
        ]
        return cls(children=children, row_align=row_align, position=(lineno, lineno))


@autodoc
@attr.s(slots=True, kw_only=True)
class Table(BlockToken):
    """Table token."""

    children: ListType[TableRow] = attr.ib(
        repr=lambda c: str(len(c)), metadata={"doc": "Child tokens list"}
    )
    header: Optional[TableRow] = attr.ib(metadata={"doc": "The header row"})
    column_align: list = attr.ib(
        metadata={
            "doc": "align options for columns (left=None (default), center=0, right=1)"
        }
    )
    position: Tuple[int, int] = attr.ib(
        metadata={"doc": "Line position in source text (start, end)"}
    )

    @staticmethod
    def split_delimiter(delimiter: str):
        """Helper function; returns a list of align options.

        :param delimiter: e.g.: `| :--- | :---: | ---: |`
        :return: a list of align options (None, 0 or 1).
        """
        return re.findall(r":?---+:?", delimiter)

    @staticmethod
    def parse_align(column):
        """Helper function; returns align option from cell content.

        :return:
            None if align = left;
            0    if align = center;
            1    if align = right.
        """
        return (0 if column[0] == ":" else 1) if column[-1] == ":" else None

    @staticmethod
    def start(line):
        return "|" in line

    @classmethod
    def read(cls, lines):
        start_line = lines.lineno + 1
        lines.anchor()
        line_buffer = [next(lines)]
        while lines.peek() is not None and "|" in lines.peek():
            line_buffer.append(next(lines))
        if len(line_buffer) < 2 or "---" not in line_buffer[1]:
            lines.reset()
            return None

        if "---" in line_buffer[1]:
            column_align = [
                cls.parse_align(column)
                for column in cls.split_delimiter(line_buffer[1])
            ]
            header = TableRow.read(line_buffer[0], column_align, lineno=start_line)
            children = [
                TableRow.read(line, column_align, lineno=start_line + i)
                for i, line in enumerate(line_buffer[2:], 2)
            ]
        else:
            column_align = [None]
            header = None
            children = [
                TableRow.read(line, lineno=start_line + i)
                for i, line in enumerate(line_buffer)
            ]
        return cls(
            children=children,
            column_align=column_align,
            header=header,
            position=(start_line, lines.lineno),
        )



"""
LaTeX renderer for mistletoe.
"""
from mistletoe import block_tokens, block_tokens_ext, span_tokens, span_tokens_ext
from mistletoe.renderers.base import BaseRenderer


class LaTeXRenderer(BaseRenderer):

    default_block_tokens = (
        # block_tokens.HTMLBlock,
        block_tokens.BlockCode,
        block_tokens.Heading,
        block_tokens.Quote,
        block_tokens.CodeFence,
        block_tokens.ThematicBreak,
        block_tokens.List,
        block_tokens_ext.Table,
        block_tokens.LinkDefinition,
        block_tokens.Paragraph,
    )

    default_span_tokens = (
        span_tokens.EscapeSequence,
        # span_tokens.HTMLSpan,
        span_tokens.AutoLink,
        span_tokens.CoreTokens,
        span_tokens_ext.Math,
        span_tokens_ext.Strikethrough,
        span_tokens.InlineCode,
        span_tokens.LineBreak,
        span_tokens.RawText,
    )

    def __init__(self, find_blocks=None, find_spans=None):
        """Initialise the renderer

        :param find_blocks: override the default block tokens (classes or class paths)
        :param find_spans: override the default span tokens (classes or class paths)
        """
        self.packages = {}
        super().__init__(find_blocks=find_blocks, find_spans=find_spans)

    def render_strong(self, token):
        return "\\textbf{{{}}}".format(self.render_inner(token))

    def render_emphasis(self, token):
        return "\\textit{{{}}}".format(self.render_inner(token))

    def render_inline_code(self, token):
        return "\\verb|{}|".format(self.render_inner(token))

    def render_strikethrough(self, token):
        self.packages["ulem"] = ["normalem"]
        return "\\sout{{{}}}".format(self.render_inner(token))

    def render_image(self, token):
        self.packages["graphicx"] = []
        return "\n\\includegraphics{{{}}}\n".format(token.src)

    def render_link(self, token):
        self.packages["hyperref"] = []
        template = "\\href{{{target}}}{{{inner}}}"
        inner = self.render_inner(token)
        return template.format(target=token.target, inner=inner)

    def render_auto_link(self, token):
        self.packages["hyperref"] = []
        return "\\url{{{}}}".format(token.target)

    @staticmethod
    def render_math(token):
        return token.content

    def render_escape_sequence(self, token):
        return self.render_inner(token)

    def render_raw_text(self, token, escape=True):
        return (
            (
                token.content.replace("$", "\\$")
                .replace("#", "\\#")
                .replace("{", "\\{")
                .replace("}", "\\}")
                .replace("&", "\\&")
            )
            if escape
            else token.content
        )

    def render_heading(self, token):
        inner = self.render_inner(token)
        if token.level == 1:
            return "\n\\section{{{}}}\n".format(inner)
        elif token.level == 2:
            return "\n\\subsection{{{}}}\n".format(inner)
        return "\n\\subsubsection{{{}}}\n".format(inner)

    def render_quote(self, token):
        self.packages["csquotes"] = []
        template = "\\begin{{displayquote}}\n{inner}\\end{{displayquote}}\n"
        return template.format(inner=self.render_inner(token))

    def render_paragraph(self, token):
        return "\n{}\n".format(self.render_inner(token))

    def render_block_code(self, token):
        self.packages["listings"] = []
        template = "\n\\begin{{lstlisting}}[language={}]\n" "{}" "\\end{{lstlisting}}\n"
        inner = self.render_raw_text(token.children[0], False)
        return template.format(token.language, inner)

    def render_list(self, token):
        self.packages["listings"] = []
        template = "\\begin{{{tag}}}\n{inner}\\end{{{tag}}}\n"
        tag = "enumerate" if token.start_at is not None else "itemize"
        inner = self.render_inner(token)
        return template.format(tag=tag, inner=inner)

    def render_list_item(self, token):
        inner = self.render_inner(token)
        return "\\item {}\n".format(inner)

    def render_table(self, token):
        def render_align(column_align):
            if column_align != [None]:
                cols = [get_align(col) for col in token.column_align]
                return "{{{}}}".format(" ".join(cols))
            else:
                return ""

        def get_align(col):
            if col is None:
                return "l"
            elif col == 0:
                return "c"
            elif col == 1:
                return "r"
            raise RuntimeError("Unrecognized align option: " + col)

        template = "\\begin{{tabular}}{align}\n" "{inner}" "\\end{{tabular}}\n"
        if hasattr(token, "header"):
            head_template = "{inner}\\hline\n"
            head_inner = self.render_table_row(token.header)
            head_rendered = head_template.format(inner=head_inner)
        else:
            head_rendered = ""
        inner = self.render_inner(token)
        align = render_align(token.column_align)
        return template.format(inner=head_rendered + inner, align=align)

    def render_table_row(self, token):
        cells = [self.render(child) for child in token.children]
        return " & ".join(cells) + " \\\\\n"

    def render_table_cell(self, token):
        return self.render_inner(token)

    @staticmethod
    def render_thematic_break(token):
        return "\\hrulefill\n"

    @staticmethod
    def render_line_break(token):
        return "\n" if token.soft else "\\newline\n"

    def render_packages(self):
        pattern = "\\usepackage{options}{{{package}}}\n"
        return "".join(
            pattern.format(options=options or "", package=package)
            for package, options in self.packages.items()
        )

    def render_document(self, token):
        template = (
            "\\documentclass{{article}}\n"
            "{packages}"
            "\\begin{{document}}\n"
            "{inner}"
            "\\end{{document}}\n"
        )
        self.link_definitions.update(token.link_definitions)
        return template.format(
            inner=self.render_inner(token), packages=self.render_packages()
        )



import pytest

from mistletoe.base_elements import serialize_tokens
from mistletoe.block_tokenizer import tokenize_main
from contrib.scheme import Scheme


@pytest.mark.parametrize(
    "name,source",
    [
        ("basic", [" (lambda (arg) (+ arg 1)\n"]),
        ("operators", ['(or (and "zero" nil "never") "James")\n']),
    ],
)
def test_tokenize(name, source, data_regression):
    with Scheme():
        data_regression.check(
            serialize_tokens(tokenize_main(source), as_dict=True),
            basename=f"test_tokenize_{name}",
        )


@pytest.mark.parametrize(
    "name,source,result",
    [
        ("basic", ["(define x (* 2 21))\n", "x\n"], 42),
        ("operators", ["(or (and 1 and 0))\n"], False),
    ],
)
def test_render(name, source, result, file_regression):
    with Scheme() as renderer:
        token = tokenize_main(source)[0]
        assert renderer.render(token) == result



"""
Block-level tokenizer for mistletoe.
"""
from mistletoe.base_elements import SpanContainer
from mistletoe.parse_context import get_parse_context


class FileWrapper:
    def __init__(self, lines, start_line=0):
        self.lines = lines if isinstance(lines, list) else list(lines)
        self._index = -1
        self._anchor = 0
        self.start_line = start_line

    @property
    def lineno(self):
        return self.start_line + self._index + 1

    def __next__(self):
        if self._index + 1 < len(self.lines):
            self._index += 1
            return self.lines[self._index]
        raise StopIteration

    def __iter__(self):
        return self

    def __repr__(self):
        return repr(self.lines[self._index + 1 :])

    def anchor(self):
        self._anchor = self._index

    def reset(self):
        self._index = self._anchor

    def peek(self):
        if self._index + 1 < len(self.lines):
            return self.lines[self._index + 1]
        return None

    def backstep(self):
        if self._index != -1:
            self._index -= 1


def tokenize_main(
    iterable, token_types=None, start_line=0, expand_spans=True, store_definitions=False
):
    """Searches for token_types in iterable.

    Args:
        iterable (list): user input lines to be parsed.
        token_types (list): a list of block-level token constructors.

    Returns:
        block-level token instances.
    """
    if token_types is None:
        token_types = get_parse_context().block_tokens
    tokens = tokenize_block(
        iterable,
        token_types=token_types,
        start_line=start_line,
        store_definitions=store_definitions,
    )
    if expand_spans:
        for token in tokens:
            for result in list(token.walk(include_self=True)):
                if isinstance(result.node.children, SpanContainer):
                    result.node.children = result.node.children.expand()
    return tokens


def tokenize_block(iterable, token_types=None, start_line=0, store_definitions=False):
    """Returns a list of parsed tokens."""
    if token_types is None:
        token_types = get_parse_context().block_tokens
    lines = FileWrapper(iterable, start_line)
    parsed_tokens = ParseBuffer()
    line = lines.peek()
    while line is not None:
        for token_type in token_types:
            if token_type.start(line):
                token = token_type.read(lines)
                if token is not None:
                    if store_definitions or token.name != "LinkDefinition":
                        parsed_tokens.append(token)
                    break
        else:  # unmatched newlines
            next(lines)
            parsed_tokens.loose = True
        line = lines.peek()
    return parsed_tokens


class ParseBuffer(list):
    """
    A wrapper around builtin list,
    so that setattr(list, 'loose') is legal.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.loose = False



from importlib import import_module

from setuptools import find_packages, setup


setup(
    name="mistletoe-ebp",
    version=import_module("mistletoe").__version__,
    description="A fast, extensible Markdown parser in pure Python.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ExecutableBookProject/mistletoe-ebp",
    project_urls={"Documentation": "https://mistletoe-ebp.readthedocs.io"},
    author="Chris Sewell",
    author_email="chrisj_sewell@hotmail.com",
    license="MIT",
    packages=find_packages(),
    entry_points={"console_scripts": ["mistletoe = mistletoe.__main__:main"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup",
    ],
    keywords="markdown lexer parser development",
    python_requires="~=3.6",
    install_requires=["attrs~=19.3"],
    extras_require={
        "code_style": ["flake8<3.8.0,>=3.7.0", "black==19.10b0", "pre-commit==1.17.0"],
        "testing": ["coverage", "pytest>=3.6,<4", "pytest-cov", "pytest-regressions"],
        "rtd": ["sphinx>=2,<3", "myst-parser~=0.4", "pyyaml"],
        "benchmark": ["commonmark~=0.9.1", "markdown~=3.2", "mistune~=0.8.4"],
    },
    zip_safe=False,
)



"""
Built-in span-level token classes.
"""
import re
from typing import Pattern, Tuple

import attr

from mistletoe import nested_tokenizer
from mistletoe.base_elements import SpanToken
from mistletoe.parse_context import get_parse_context
from mistletoe.attr_doc import autodoc

"""
Tokens to be included in the parsing process, in the order specified.
"""
__all__ = [
    "EscapeSequence",
    "AutoLink",
    "CoreTokens",
    "InlineCode",
    "LineBreak",
    "RawText",
]


class CoreTokens(SpanToken):
    precedence = 3

    @classmethod
    def read(cls, match: Pattern):
        # TODO this needs to be made more general (so tokens can be in diffent modules)
        return globals()[match.type].read(match)

    @classmethod
    def find(cls, string):
        return nested_tokenizer.find_nested_tokenizer(string)


class Strong(SpanToken):
    """
    Strong tokens. ("**some text**")

    :ivar content: raw string content of the token
    :ivar children: list of child tokens
    """


class Emphasis(SpanToken):
    """
    Emphasis tokens. ("*some text*")

    :ivar content: raw string content of the token
    :ivar children: list of child tokens
    """


@autodoc
@attr.s(kw_only=True, slots=True)
class InlineCode(SpanToken):
    """
    Inline code tokens. ("`some code`")
    """

    pattern = re.compile(r"(?<!\\|`)(?:\\\\)*(`+)(?!`)(.+?)(?<!`)\1(?!`)", re.DOTALL)
    parse_inner = False
    parse_group = 2

    children = attr.ib(
        repr=False, metadata={"doc": "a single RawText node for alternative text."}
    )
    position: Tuple[int, int] = attr.ib(
        default=None,
        repr=False,
        metadata={"doc": "Line position in source text (start, end)"},
    )

    @classmethod
    def read(cls, match: Pattern):
        content = match.group(cls.parse_group)
        return cls(children=(RawText(" ".join(re.split("[ \n]+", content.strip()))),))

    @classmethod
    def find(cls, string):
        matches = get_parse_context().nesting_matches.pop("InlineCode", [])
        return matches


@autodoc
@attr.s(kw_only=True, slots=True)
class Image(SpanToken):
    """
    Image tokens, with inline targets: "![alt](src "title")".
    """

    src: str = attr.ib(metadata={"doc": "image source"})
    title: str = attr.ib(default=None, metadata={"doc": "image title"})
    children = attr.ib(factory=list, repr=False, metadata={"doc": "alternative text."})
    position: Tuple[int, int] = attr.ib(
        default=None,
        repr=False,
        metadata={"doc": "Line position in source text (start, end)"},
    )

    @classmethod
    def read(cls, match: Pattern):
        return cls(src=match.group(2).strip(), title=match.group(3))


@autodoc
@attr.s(kw_only=True, slots=True)
class Link(SpanToken):
    """
    Link tokens, with inline targets: "[name](target)"
    """

    target: str = attr.ib(metadata={"doc": "link target"})
    title: str = attr.ib(default=None, metadata={"doc": "link title"})
    children = attr.ib(factory=list, repr=False, metadata={"doc": "link text."})
    position: Tuple[int, int] = attr.ib(
        default=None,
        repr=False,
        metadata={"doc": "Line position in source text (start, end)"},
    )

    @classmethod
    def read(cls, match: Pattern):
        return cls(
            target=EscapeSequence.strip(match.group(2).strip()),
            title=EscapeSequence.strip(match.group(3)),
        )


@autodoc
@attr.s(kw_only=True, slots=True)
class AutoLink(SpanToken):
    """
    Autolink tokens. ("<http://www.google.com>")
    """

    pattern = re.compile(
        r"(?<!\\)(?:\\\\)*<([A-Za-z][A-Za-z0-9+.-]{1,31}:[^ <>]*?|[A-Za-z0-9.!#$%&'*+/=?^_`{|}~-]+@[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?(?:\.[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*)>"  # noqa: E501
    )
    parse_inner = False

    target: str = attr.ib(metadata={"doc": "link target"})
    mailto: bool = attr.ib(metadata={"doc": "if the link is an email"})
    children = attr.ib(
        repr=False, metadata={"doc": "a single RawText node for alternative text."}
    )
    position: Tuple[int, int] = attr.ib(
        default=None,
        repr=False,
        metadata={"doc": "Line position in source text (start, end)"},
    )

    @classmethod
    def read(cls, match: Pattern):
        content = match.group(cls.parse_group)
        return cls(
            children=(RawText(content),),
            target=content,
            mailto="@" in content and "mailto" not in content.casefold(),
        )


@autodoc
@attr.s(kw_only=True, slots=True)
class EscapeSequence(SpanToken):
    """
    Escape sequences. ("\\*")

    Attributes:
        children (iterator): a single RawText node for alternative text.
    """

    pattern = re.compile(r"\\([!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~])")
    parse_inner = False
    precedence = 2

    children = attr.ib(
        repr=False, metadata={"doc": "a single RawText node for alternative text."}
    )
    position: Tuple[int, int] = attr.ib(
        default=None,
        repr=False,
        metadata={"doc": "Line position in source text (start, end)"},
    )

    @classmethod
    def read(cls, match: Pattern):
        return cls(children=(RawText(match.group(cls.parse_group)),))

    @classmethod
    def strip(cls, string):
        return cls.pattern.sub(r"\1", string)


@autodoc
@attr.s(kw_only=True, slots=True)
class LineBreak(SpanToken):
    """
    Hard or soft line breaks.
    """

    pattern = re.compile(r"( *|\\)\n")
    parse_inner = False
    parse_group = 0

    content: bool = attr.ib(default="", repr=False, metadata={"doc": "raw content."})
    soft: bool = attr.ib(metadata={"doc": "if the break is soft or hard."})
    position: Tuple[int, int] = attr.ib(
        default=None,
        repr=False,
        metadata={"doc": "Line position in source text (start, end)"},
    )

    @classmethod
    def read(cls, match: Pattern):
        content = match.group(1)
        return cls(soft=not content.startswith(("  ", "\\")))


@autodoc
@attr.s(slots=True)
class RawText(SpanToken):
    """
    Raw text. A leaf node.

    RawText is the only token that accepts a string for its `read` method,
    instead of a match object. Also, all recursions should bottom out here.
    """

    content: bool = attr.ib(
        repr=False, metadata={"doc": "raw string content of the token"}
    )
    position: Tuple[int, int] = attr.ib(
        default=None,
        repr=False,
        metadata={"doc": "Line position in source text (start, end)"},
    )

    @classmethod
    def read(cls, content: str):
        return cls(content=content)


_tags = {
    "address",
    "article",
    "aside",
    "base",
    "basefont",
    "blockquote",
    "body",
    "caption",
    "center",
    "col",
    "colgroup",
    "dd",
    "details",
    "dialog",
    "dir",
    "div",
    "dl",
    "dt",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "frame",
    "frameset",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "head",
    "header",
    "hr",
    "html",
    "iframe",
    "legend",
    "li",
    "link",
    "main",
    "menu",
    "menuitem",
    "meta",
    "nav",
    "noframes",
    "ol",
    "optgroup",
    "option",
    "p",
    "param",
    "section",
    "source",
    "summary",
    "table",
    "tbody",
    "td",
    "tfoot",
    "th",
    "thead",
    "title",
    "tr",
    "track",
    "ul",
}

_tag = r"[A-Za-z][A-Za-z0-9-]*"
_attrs = r'(?:\s+[A-Za-z_:][A-Za-z0-9_.:-]*(?:\s*=\s*(?:[^ "\'=<>`]+|\'[^\']*?\'|"[^\"]*?"))?)*'  # noqa: E501

_open_tag = r"(?<!\\)<" + _tag + _attrs + r"\s*/?>"
_closing_tag = r"(?<!\\)</" + _tag + r"\s*>"
_comment = r"(?<!\\)<!--(?!>|->)(?:(?!--).)+?(?<!-)-->"
_instruction = r"(?<!\\)<\?.+?\?>"
_declaration = r"(?<!\\)<![A-Z].+?>"
_cdata = r"(?<!\\)<!\[CDATA.+?\]\]>"


class HTMLSpan(SpanToken):
    """
    Span-level HTML tokens.

    :ivar content: raw string content of the token
    :ivar children: list of child tokens
    """

    pattern = re.compile(
        "|".join(
            [_open_tag, _closing_tag, _comment, _instruction, _declaration, _cdata]
        ),
        re.DOTALL,
    )
    parse_inner = False
    parse_group = 0



import re
import sys
import json
from mistletoe import markdown
from traceback import print_tb
from argparse import ArgumentParser

KNOWN = ()


def run_tests(
    test_entries, start=None, end=None, quiet=False, verbose=False, known=False
):
    global KNOWN
    if known:
        print("ignoring tests:", ", ".join(map(str, KNOWN)) + "\n")
    start = start or 0
    end = end or sys.maxsize
    results = [
        run_test(test_entry, quiet)
        for test_entry in test_entries
        if test_entry["example"] >= start
        and test_entry["example"] <= end
        and (not known or test_entry["example"] not in KNOWN)
    ]
    if verbose:
        print_failure_in_sections(results)
    fails = len(list(filter(lambda x: not x[0], results)))
    if fails:
        print("failed:", fails)
        print(" total:", len(results))
    else:
        print("All tests passing.")
    return not fails


def run_test(test_entry, quiet=False):
    test_case = test_entry["markdown"].splitlines(keepends=True)
    try:
        output = markdown(test_case)
        success = test_entry["html"] == output
        if not success and not quiet:
            print_test_entry(test_entry, output)
        return success, test_entry["section"]
    except Exception as exception:
        if not quiet:
            print_exception(exception, test_entry)
        return False, test_entry["section"]


def load_tests(specfile):
    with open(specfile, "r") as fin:
        return json.load(fin)


def locate_section(section, tests):
    start = None
    end = None
    for test in tests:
        if re.search(section, test["section"], re.IGNORECASE):
            if start is None:
                start = test["example"]
        elif start is not None and end is None:
            end = test["example"] - 1
            return start, end
    if start:
        return start, tests[-1]["example"] - 1
    raise RuntimeError("Section '{}' not found, aborting.".format(section))


def print_exception(exception, test_entry):
    print_test_entry(test_entry, "-- exception --", fout=sys.stderr)
    print(exception.__class__.__name__ + ":", exception, file=sys.stderr)
    print("Traceback: ", file=sys.stderr)
    print_tb(exception.__traceback__)


def print_test_entry(test_entry, output, fout=sys.stdout):
    print("example: ", repr(test_entry["example"]), file=fout)
    print("markdown:", repr(test_entry["markdown"]), file=fout)
    print("html:    ", repr(test_entry["html"]), file=fout)
    print("output:  ", repr(output), file=fout)
    print(file=fout)


def print_failure_in_sections(results):
    section = results[0][1]
    failed = 0
    total = 0
    for result in results:
        if section != result[1]:
            if failed:
                section_str = "Failed in section '{}':".format(section)
                result_str = "{:>3} / {:>3}".format(failed, total)
                print("{:70} {}".format(section_str, result_str))
            section = result[1]
            failed = 0
            total = 0
        if not result[0]:
            failed += 1
        total += 1
    if failed:
        section_str = "Failed in section '{}':".format(section)
        result_str = "{:>3} / {:>3}".format(failed, total)
        print("{:70} {}".format(section_str, result_str))
    print()


def main():
    parser = ArgumentParser(description="Custom script for running Commonmark tests.")
    parser.add_argument(
        "start",
        type=int,
        nargs="?",
        default=None,
        help="Run tests starting from this position.",
    )
    parser.add_argument(
        "end", type=int, nargs="?", default=None, help="Run tests until this position."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Output failure count in every section.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="Suppress failed test entry output.",
    )
    parser.add_argument(
        "-s",
        "--section",
        dest="section",
        default=None,
        help="Only run tests in specified section.",
    )
    parser.add_argument(
        "-f",
        "--file",
        dest="tests",
        type=load_tests,
        default="test/commonmark/commonmark.json",
        help="Specify alternative specfile to run.",
    )
    parser.add_argument(
        "-n",
        "--ignore-known",
        dest="known",
        action="store_true",
        help="Ignore tests entries that are known to fail.",
    )
    args = parser.parse_args()

    start = args.start
    end = args.end
    verbose = args.verbose
    quiet = args.quiet
    tests = args.tests
    known = args.known
    if args.section is not None:
        start, end = locate_section(args.section, tests)

    if not run_tests(tests, start, end, quiet, verbose, known):
        sys.exit(1)


if __name__ == "__main__":
    main()



"""This module provides a container for global variables of a single parse.

It uses the `threading.local` object to ensure that global variables
are not changed by different threads.
"""
from copy import deepcopy
from importlib import import_module
from threading import local

THREAD = local()


class ParseContext:
    def __init__(self, find_blocks=None, find_spans=None, link_definitions=None):
        """A class to contain context for a single parse.

        :param find_blocks: a list of block tokens to use during the parse.
            If None the standard blocks will be used from `block_tokens.__all__`
        :param find_spans: a list of span tokens to use during the parse.
            If None the standard blocks will be used from `span_tokens.__all__`
        :param link_definitions: a dict of link definitons, obtained from `[def]: link`
        :param nesting_matches: a dict of matches recorded from `find_nested_tokenizer`
        """
        if link_definitions is None:
            self.link_definitions = {}
        else:
            self.link_definitions = link_definitions
        if find_blocks is not None:
            self.block_tokens = tokens_from_classes(find_blocks)
        else:
            from mistletoe.renderers.base import BaseRenderer

            self.block_tokens = list(BaseRenderer.default_block_tokens)
        if find_spans is not None:
            self.span_tokens = tokens_from_classes(find_spans)
        else:
            from mistletoe.renderers.base import BaseRenderer

            self.span_tokens = list(BaseRenderer.default_span_tokens)

        self.nesting_matches = {}

    def copy(self):
        return deepcopy(self)


def get_parse_context(reset=False) -> ParseContext:
    """Return the current `ParseContext`."""
    global THREAD
    if reset:
        THREAD.context = ParseContext()
    else:
        try:
            THREAD.context
        except AttributeError:
            THREAD.context = ParseContext()
    return THREAD.context


def set_parse_context(parse_context):
    """Set an existing `ParseContext`."""
    global THREAD
    THREAD.context = parse_context


def tokens_from_module(module):
    """
    Helper method; takes a module and returns a list of all token classes
    specified in module.__all__.
    Useful when custom tokens are defined in single module.
    """
    return [getattr(module, name) for name in module.__all__]


def tokens_from_classes(classes):
    """
    Helper method; take a list of classes and/or class paths
    (e.g. `mistletoe.span_tokens.Math`) and return the loaded classes.
    """
    return [
        getattr(import_module(".".join(cls.split(".")[:-1])), cls.split(".")[-1])
        if isinstance(cls, str)
        else cls
        for cls in classes
    ]
